import sys

import click

from iseq.gff import read as read_gff


@click.command()
@click.argument("gff_file", type=click.File("r"))
@click.option(
    "--max-e-value",
    type=float,
    default=None,
    help="Filter out items for which E-value > --max-e-evalue.",
)
def gff_filter(gff_file, max_e_value):
    """
    Filter out items from a GFF_FILE file.

    The resulting file will be written to the standard output.
    """

    gff = read_gff(gff_file)

    kwargs = {}
    if max_e_value is not None:
        kwargs["max_e_value"] = float(max_e_value)

    gff = gff.filter(**kwargs)

    gff.write_file(sys.stdout)



from hmmer_reader import open_hmmer
from imm import Sequence
from imm.testing import assert_allclose
from nmm import IUPACAminoAlphabet, RNAAlphabet

from iseq.codon_table import CodonTable
from iseq.example import example_filepath
from iseq.hmmer_model import HMMERModel
from iseq.protein import create_profile


def test_protein_profile_frame1():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        plain_model = reader.read_model()
        hmmer = create_profile(HMMERModel(plain_model), RNAAlphabet())

    rna_abc = hmmer.alphabet
    most_likely_rna_seq = b"CCU GGU AAA GAA GAU AAU AAC AAA".replace(b" ", b"")
    most_likely_seq = Sequence.create(most_likely_rna_seq, rna_abc)
    r = hmmer.search(most_likely_seq).results[0]

    assert_allclose(r.loglikelihood, 125.83363182422178)
    frags = r.fragments
    assert len(frags) == 1
    frag = frags[0]
    assert frag.homologous
    assert bytes(frag.sequence) == bytes(most_likely_seq)
    desired = "('CCU', '<M1,3>'),('GGU', '<M2,3>'),('AAA', '<M3,3>')"
    assert str(frag)[:53] == desired


def test_protein_profile_frame2():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        plain_model = reader.read_model()
        hmmer = create_profile(HMMERModel(plain_model), RNAAlphabet(), epsilon=0.1)

    rna_abc = hmmer.alphabet
    rna_seq = b"AAA AAA AAA CCU GGU AAA GAA GAU AAU AAC AAA"
    rna_seq = rna_seq.replace(b" ", b"")
    seq = Sequence.create(rna_seq, rna_abc)

    r = hmmer.search(seq).results[0]
    assert_allclose(r.loglikelihood, 168.23071232889802)
    frags = r.fragments
    assert len(frags) == 2
    assert not frags[0].homologous
    assert bytes(frags[0].sequence) == b"AAAAAAAAA"
    assert frags[1].homologous
    assert bytes(frags[1].sequence) == b"CCUGGUAAAGAAGAUAAUAACAAA"


def test_protein_profile_frame3():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        plain_model = reader.read_model()
        hmmer = create_profile(HMMERModel(plain_model), RNAAlphabet(), epsilon=0.0)

    rna_abc = hmmer.alphabet
    rna_seq = b"CCU GGU AAA GAA GAU AAU AAC AAA"
    rna_seq = rna_seq.replace(b" ", b"")
    seq = Sequence.create(rna_seq, rna_abc)

    r = hmmer.search(seq).results[0]
    frags = r.fragments
    assert len(frags) == 1
    assert frags[0].homologous
    assert bytes(frags[0].sequence) == b"CCUGGUAAAGAAGAUAAUAACAAA"


def test_protein_profile_frame4():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        plain_model = reader.read_model()
        hmmer = create_profile(HMMERModel(plain_model), RNAAlphabet(), epsilon=0.0)

    rna_abc = hmmer.alphabet
    rna_seq = b"CCUU GGU AAA GAA GAU AAU AAC AAA"
    rna_seq = rna_seq.replace(b" ", b"")
    seq = Sequence.create(rna_seq, rna_abc)

    r = hmmer.search(seq).results[0]
    frags = r.fragments
    assert len(frags) == 0


def test_protein_profile_frame5():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        plain_model = reader.read_model()
        hmmer = create_profile(HMMERModel(plain_model), RNAAlphabet(), epsilon=0.00001)

    rna_abc = hmmer.alphabet
    rna_seq = b"CCUU GGU AAA GAA GAU AAU AAC AAA"
    rna_seq = rna_seq.replace(b" ", b"")
    seq = Sequence.create(rna_seq, rna_abc)

    r = hmmer.search(seq).results[0]
    frags = r.fragments
    assert len(frags) == 1
    assert frags[0].homologous
    assert bytes(frags[0].sequence) == b"CCUUGGUAAAGAAGAUAAUAACAAA"


def test_protein_profile_frame6():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        plain_model = reader.read_model()
        hmmer = create_profile(HMMERModel(plain_model), RNAAlphabet(), epsilon=0.00001)

    rna_abc = hmmer.alphabet
    rna_seq = b"CCUU GGU AAA GAA GAU AAU AAC AAA GAA GAA CCU GGU AAA GAA GAU AAU AAC AAA GAA GAA GA"
    rna_seq = rna_seq.replace(b" ", b"")
    seq = Sequence.create(rna_seq, rna_abc)

    r = hmmer.search(seq).results[0]
    frags = r.fragments
    assert len(frags) == 4
    assert frags[0].homologous
    assert bytes(frags[0].sequence) == b"CCUUGGUAAAGAAGAUAAUAACAAA"
    assert not frags[1].homologous
    assert bytes(frags[1].sequence) == b"GAAGAA"
    assert frags[2].homologous
    assert bytes(frags[2].sequence) == b"CCUGGUAAAGAAGAUAAUAACAAA"
    assert not frags[3].homologous
    assert bytes(frags[3].sequence) == b"GAAGAAGA"

    hmmer.multiple_hits = False
    r = hmmer.search(seq).results[0]
    frags = r.fragments
    assert_allclose(r.loglikelihood, 1445.0314253859958)
    assert len(frags) == 3
    assert not frags[0].homologous
    assert bytes(frags[0].sequence) == b"CCUUGGUAAAGAAGAUAAUAACAAAGAAGAA"
    assert frags[1].homologous
    assert bytes(frags[1].sequence) == b"CCUGGUAAAGAAGAUAAUAACAAA"
    assert not frags[2].homologous
    assert bytes(frags[2].sequence) == b"GAAGAAGA"


def test_protein_profile_codons():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        plain_model = reader.read_model()
        hmmer = create_profile(HMMERModel(plain_model), RNAAlphabet(), epsilon=0.1)

    rna_abc = hmmer.alphabet
    amino_abc = IUPACAminoAlphabet()
    gcode = CodonTable(rna_abc, amino_abc)

    rna_seq = b"AAGA AAA AAA CCU GGU AAA GAA GAU AAU AAC AAA G"
    rna_seq = rna_seq.replace(b" ", b"")
    seq = Sequence.create(rna_seq, rna_abc)

    r = hmmer.search(seq).results[0]
    assert_allclose(r.loglikelihood, 175.35113397356454)
    frags = r.fragments
    cfrags = [f.decode() for f in frags]
    aafrags = [f.decode(gcode) for f in cfrags]

    assert len(frags) == 2
    assert len(cfrags) == 2
    assert len(aafrags) == 2

    assert not frags[0].homologous
    assert not cfrags[0].homologous
    assert not aafrags[0].homologous

    assert bytes(frags[0].sequence) == b"AAGAAAAAAA"
    assert bytes(cfrags[0].sequence) == b"AAGAAAAAA"
    assert bytes(aafrags[0].sequence) == b"KKK"

    items = list(iter(frags[0]))
    citems = list(iter(cfrags[0]))
    aaitems = list(iter(aafrags[0]))

    assert bytes(items[0].sequence) == b""
    assert str(items[0].step) == "<S,0>"
    assert bytes(citems[0].sequence) == b""
    assert str(citems[0].step) == "<S,0>"
    assert bytes(aaitems[0].sequence) == b""
    assert str(aaitems[0].step) == "<S,0>"

    assert bytes(items[1].sequence) == b"AAG"
    assert str(items[1].step) == "<N,3>"
    assert bytes(citems[1].sequence) == b"AAG"
    assert str(citems[1].step) == "<N,3>"
    assert bytes(aaitems[1].sequence) == b"K"
    assert str(aaitems[1].step) == "<N,1>"

    assert bytes(items[2].sequence) == b"AAAA"
    assert str(items[2].step) == "<N,4>"
    assert bytes(citems[2].sequence) == b"AAA"
    assert str(citems[2].step) == "<N,3>"
    assert bytes(aaitems[2].sequence) == b"K"
    assert str(aaitems[2].step) == "<N,1>"

    assert bytes(items[3].sequence) == b"AAA"
    assert str(items[3].step) == "<N,3>"
    assert bytes(citems[3].sequence) == b"AAA"
    assert str(citems[3].step) == "<N,3>"
    assert bytes(aaitems[3].sequence) == b"K"
    assert str(aaitems[3].step) == "<N,1>"

    assert bytes(items[4].sequence) == b""
    assert str(items[4].step) == "<B,0>"
    assert bytes(citems[4].sequence) == b""
    assert str(citems[4].step) == "<B,0>"
    assert bytes(aaitems[4].sequence) == b""
    assert str(aaitems[4].step) == "<B,0>"

    assert frags[1].homologous
    assert cfrags[1].homologous
    assert aafrags[1].homologous
    assert bytes(frags[1].sequence) == b"CCUGGUAAAGAAGAUAAUAACAAAG"
    assert bytes(cfrags[1].sequence) == b"CCUGGUAAAGAAGAUAAUAACAAG"

    assert bytes(aafrags[1].sequence) == b"PGKEDNNK"

    items = list(iter(frags[1]))
    citems = list(iter(cfrags[1]))
    aaitems = list(iter(aafrags[1]))

    assert bytes(items[0].sequence) == b"CCU"
    assert str(items[0].step) == "<M1,3>"
    assert bytes(citems[0].sequence) == b"CCU"
    assert str(citems[0].step) == "<M1,3>"
    assert bytes(aaitems[0].sequence) == b"P"
    assert str(aaitems[0].step) == "<M1,1>"

    assert bytes(items[7].sequence) == b"AAAG"
    assert str(items[7].step) == "<M8,4>"
    assert bytes(citems[7].sequence) == b"AAG"
    assert str(citems[7].step) == "<M8,3>"
    assert bytes(aaitems[7].sequence) == b"K"
    assert str(aaitems[7].step) == "<M8,1>"



import click
import nmm
from fasta_reader import FASTAWriter, open_fasta

from iseq.codon_table import CodonTable
from iseq.gencode import GeneticCode


@click.command()
@click.argument("fasta", type=click.File("r"))
@click.option(
    "--output",
    type=click.File("w", lazy=True),
    help="Save results to OUTPUT (FASTA format).",
    default="-",
)
@click.option(
    "--seed",
    type=int,
    help="Random seed used to initialize the pseudo-random number generator.",
    default=0,
)
@click.option(
    "--transl-table",
    type=int,
    help="Translation table number. Defaults to 1 (The Standard Code).",
    default=1,
)
@click.option(
    "--alph",
    type=click.Choice(["dna", "rna"]),
    help="Nucleotide alphabet.",
    default="dna",
)
def amino_decode(fasta, output, seed: int, transl_table: int, alph: str):
    from numpy.random import RandomState

    owriter = FASTAWriter(output, 60)

    if alph == "dna":
        base_abc = nmm.DNAAlphabet()
    else:
        base_abc = nmm.RNAAlphabet()

    amino_abc = nmm.IUPACAminoAlphabet()
    codon_table = CodonTable(base_abc, amino_abc, GeneticCode(id=transl_table))

    random = RandomState(seed)
    for target in open_fasta(fasta):
        amino_seq = target.sequence
        nucls = []
        for amino in amino_seq:
            codons = codon_table.codons(amino.encode())
            codon = random.choice(codons)
            nucls.append(codon.symbols.decode())

        owriter.write_item(target.defline, "".join(nucls))

    output.close_intelligently()



import os

from fasta_reader import open_fasta
from hmmer_reader import open_hmmer
from imm import Sequence
from imm.testing import assert_allclose
from numpy.testing import assert_equal

from iseq.example import example_filepath
from iseq.hmmer3 import create_profile
from iseq.hmmer_model import HMMERModel
from iseq.model import EntryDistr


def test_hmmer3_profile_unihit_homologous_1():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        hmmdata = HMMERModel(reader.read_model())

    hmmer = create_profile(hmmdata, entry_distr=EntryDistr.UNIFORM)

    alphabet = hmmer.alphabet
    most_likely_seq = Sequence.create(b"PGKEDNNK", alphabet)
    r = hmmer.search(most_likely_seq).results[0]

    assert_allclose(r.loglikelihood, 13.39978964458627)
    frags = r.fragments
    assert_equal(len(frags), 1)
    frag = frags[0]
    assert_equal(frag.homologous, True)
    assert_equal(bytes(frag.sequence), bytes(most_likely_seq))

    hmmer.multiple_hits = False
    r = hmmer.search(most_likely_seq).results[0]
    assert_allclose(r.loglikelihood, 13.472626968538535)
    frags = r.fragments
    assert_equal(len(frags), 1)
    frag = frags[0]
    assert_equal(frag.homologous, True)
    assert_equal(bytes(frag.sequence), bytes(most_likely_seq))


def test_hmmer3_profile_problematic1():
    with open_hmmer(example_filepath("problematic1.hmm")) as reader:
        hmmdata = HMMERModel(reader.read_model())

    prof = create_profile(hmmdata, True)

    with open_fasta(example_filepath("problematic1.fasta")) as reader:
        item = reader.read_items()[0]

    sequence = Sequence.create(item.sequence.encode(), prof.alphabet)
    r = prof.search(sequence)
    assert len(r.results) == 1
    assert_allclose(r.results[0].alt_viterbi_score, -5.103729125680717)


def test_hmmer3_profile_small_viterbi_score(tmp_path):
    os.chdir(tmp_path)
    profile = example_filepath("PF15449.6.hmm")
    fasta = example_filepath("A0ALD9.fasta")

    with open_hmmer(profile) as reader:
        hmmdata = HMMERModel(reader.read_model())

    prof = create_profile(hmmdata, True)

    with open_fasta(fasta) as reader:
        item = reader.read_items()[0]

    sequence = Sequence.create(item.sequence.encode(), prof.alphabet)
    r = prof.search(sequence)
    assert len(r.results) == 1
    assert_allclose(r.results[0].alt_viterbi_score, -18.424065160005625)


def test_hmmer3_profile_large_viterbi_score(tmp_path):
    os.chdir(tmp_path)
    profile = example_filepath("PF07476.11.hmm")
    fasta = example_filepath("A0ALD9.fasta")

    with open_hmmer(profile) as reader:
        hmmdata = HMMERModel(reader.read_model())

    prof = create_profile(hmmdata, True)

    with open_fasta(fasta) as reader:
        item = reader.read_items()[0]

    sequence = Sequence.create(item.sequence.encode(), prof.alphabet)
    r = prof.search(sequence)
    assert len(r.results) == 1
    assert_allclose(r.results[0].alt_viterbi_score, 3.480341268180834)


def test_hmmer3_profile_unihit_homologous_2():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        hmmdata = HMMERModel(reader.read_model())

    hmmer = create_profile(hmmdata, entry_distr=EntryDistr.UNIFORM)

    alphabet = hmmer.alphabet
    seq = Sequence.create(b"PGKENNK", alphabet)
    r = hmmer.search(seq).results[0]
    assert_allclose(r.loglikelihood, 4.679028352468922)
    frags = r.fragments
    assert_equal(len(frags), 1)
    frag = frags[0]
    assert_equal(frag.homologous, True)
    assert_equal(bytes(frag.sequence), bytes(seq))
    assert_equal(str(frag)[:31], "('P', '<M1,1>'),('G', '<M2,1>')")


def test_hmmer3_profile_unihit_homologous_3():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        hmmdata = HMMERModel(reader.read_model())

    hmmer = create_profile(hmmdata, entry_distr=EntryDistr.UNIFORM)

    alphabet = hmmer.alphabet
    seq = Sequence.create(b"PGKEPNNK", alphabet)
    r = hmmer.search(seq).results[0]
    assert_allclose(r.loglikelihood, 8.55629338255035)
    frags = r.fragments
    assert_equal(len(frags), 1)
    frag = frags[0]
    assert_equal(frag.homologous, True)
    assert_equal(bytes(frag.sequence), bytes(seq))


def test_hmmer3_profile_nonhomo_and_homologous():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        hmmdata = HMMERModel(reader.read_model())

    hmmer = create_profile(hmmdata, entry_distr=EntryDistr.UNIFORM)

    alphabet = hmmer.alphabet
    seq = Sequence.create(b"KKKPGKEDNNK", alphabet)
    assert_equal(hmmer.multiple_hits, True)
    r = hmmer.search(seq).results[0]
    assert_allclose(r.loglikelihood, 12.23961188080344)
    frags = r.fragments
    assert_equal(len(frags), 2)
    assert_equal(frags[0].homologous, False)
    assert_equal(bytes(frags[0].sequence), b"KKK")
    assert_equal(frags[1].homologous, True)
    assert_equal(bytes(frags[1].sequence), b"PGKEDNNK")

    hmmer.multiple_hits = False
    assert_equal(hmmer.multiple_hits, False)
    r = hmmer.search(seq).results[0]
    assert_allclose(r.loglikelihood, 12.492368705915666)
    frags = r.fragments
    assert_equal(len(frags), 2)
    assert_equal(frags[0].homologous, False)
    assert_equal(bytes(frags[0].sequence), b"KKK")
    assert_equal(frags[1].homologous, True)
    assert_equal(bytes(frags[1].sequence), b"PGKEDNNK")


def test_hmmer3_profile_multihit_homologous1():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        hmmdata = HMMERModel(reader.read_model())

    hmmer = create_profile(hmmdata, entry_distr=EntryDistr.OCCUPANCY)

    alphabet = hmmer.alphabet
    seq = Sequence.create(b"PPPPGKEDNNKDDDPGKEDNNKEEEE", alphabet)
    r = hmmer.search(seq).results[0]
    assert_allclose(r.loglikelihood, 23.396127957291384)
    frags = r.fragments
    assert_equal(len(frags), 5)
    assert_equal(frags[0].homologous, False)
    assert_equal(bytes(frags[0].sequence), b"PPP")
    assert_equal(frags[1].homologous, True)
    assert_equal(bytes(frags[1].sequence), b"PGKEDNNK")
    assert_equal(frags[3].homologous, True)
    assert_equal(bytes(frags[3].sequence), b"PGKEDNNK")
    assert_equal(frags[4].homologous, False)
    assert_equal(bytes(frags[4].sequence), b"EEEE")

    items = list(iter(frags[0]))

    assert_equal(bytes(items[0].sequence), b"")
    assert_equal(str(items[0].step), "<S,0>")
    assert_equal(bytes(items[1].sequence), b"P")
    assert_equal(str(items[1].step), "<N,1>")
    assert_equal(bytes(items[4].sequence), b"")
    assert_equal(str(items[4].step), "<B,0>")

    items = list(frags[1])

    assert_equal(bytes(items[0].sequence), b"P")
    assert_equal(str(items[0].step), "<M1,1>")
    assert_equal(bytes(items[1].sequence), b"G")
    assert_equal(str(items[1].step), "<M2,1>")
    assert_equal(bytes(items[2].sequence), b"K")
    assert_equal(str(items[2].step), "<M3,1>")
    assert_equal(bytes(items[3].sequence), b"E")
    assert_equal(str(items[3].step), "<M4,1>")
    assert_equal(bytes(items[4].sequence), b"D")
    assert_equal(str(items[4].step), "<M5,1>")
    assert_equal(bytes(items[5].sequence), b"N")
    assert_equal(str(items[5].step), "<M6,1>")
    assert_equal(bytes(items[6].sequence), b"N")
    assert_equal(str(items[6].step), "<M7,1>")
    assert_equal(bytes(items[7].sequence), b"K")
    assert_equal(str(items[7].step), "<M8,1>")

    hmmer.multiple_hits = False
    r = hmmer.search(seq).results[0]
    assert_allclose(r.loglikelihood, 10.19992887279622)
    frags = r.fragments
    assert_equal(len(frags), 3)
    assert_equal(frags[0].homologous, False)
    assert_equal(frags[1].homologous, True)
    assert_equal(bytes(frags[1].sequence), b"PGKEDNNK")
    assert_equal(frags[2].homologous, False)


def test_hmmer3_profile_window():
    filepath = example_filepath("PF03373.hmm")
    with open_hmmer(filepath) as reader:
        hmmdata = HMMERModel(reader.read_model())

    hmmer = create_profile(hmmdata, entry_distr=EntryDistr.UNIFORM, window_length=15)

    alphabet = hmmer.alphabet
    seq = Sequence.create(b"PPPPGKEDNNKDDDPGKEDNNKEEEE", alphabet)
    r = hmmer.search(seq)

    assert_allclose(r.results[0].loglikelihood, 10.400476828620292)
    frags = r.results[0].fragments
    assert_equal(len(frags), 3)
    assert_equal(frags[0].homologous, False)
    assert_equal(bytes(frags[0].sequence), b"PPP")
    assert_equal(frags[1].homologous, True)
    assert_equal(bytes(frags[1].sequence), b"PGKEDNNK")
    assert_equal(frags[2].homologous, False)
    assert_equal(bytes(frags[2].sequence), b"DDDP")

    assert_allclose(r.results[1].loglikelihood, 14.513128891939548)
    frags = r.results[1].fragments
    assert_equal(len(frags), 3)
    assert_equal(frags[0].homologous, True)
    assert_equal(bytes(frags[0].sequence), b"DNNK")
    assert_equal(frags[1].homologous, False)
    assert_equal(bytes(frags[1].sequence), b"DDD")
    assert_equal(frags[2].homologous, True)
    assert_equal(bytes(frags[2].sequence), b"PGKEDNNK")

    assert_allclose(r.results[2].loglikelihood, 10.614853720566728)
    frags = r.results[2].fragments
    assert_equal(len(frags), 2)
    assert_equal(frags[0].homologous, True)
    assert_equal(bytes(frags[0].sequence), b"PGKEDNNK")
    assert_equal(frags[1].homologous, False)
    assert_equal(bytes(frags[1].sequence), b"EEEE")



# It might be useful in the future. Not for now.
# import sys

# import click

# from iseq.gff import read as read_gff


# @click.command()
# @click.argument("gff_file", type=click.File("r"))
# def gff_dedup(gff_file):
#     """
#     Deduplicate a GFF_FILE file.

#     The deduplicated GFF_FILE will be written to the standard output.
#     """

#     gff = read_gff(gff_file)
#     gff.deduplicate()

#     gff.write_file(sys.stdout)



import os
from io import StringIO

from numpy.testing import assert_equal

from iseq.example import example_filepath
from iseq.gff import read as read_gff
from iseq.testing import assert_same_file


def test_gff_read():
    items = read_gff(example_filepath("duplicate.gff")).items()
    assert_equal(len(items), 14)
    assert_equal(items[3].seqid, "GALNBKIG_00914_ont_01_plus_strand")
    assert_equal(items[6].end, 474)


def test_gff_deduplicate(tmp_path):
    os.chdir(tmp_path)

    gff = read_gff(example_filepath("duplicate.gff"))
    gff.deduplicate()

    gff.write_file("output.gff")

    dedup_file = example_filepath("deduplicate.gff")
    assert_same_file("output.gff", dedup_file)


def test_gff_read_empty():
    file = StringIO("##gff-version 3")
    gff = read_gff(file)
    df = gff.to_dataframe()
    assert_equal(len(df), 0)



from typing import TypeVar

from imm import Alphabet, NormalState, Step

from iseq.fragment import Fragment
from iseq.model import AltModel, Node, NullModel, SpecialNode
from iseq.result import SearchResults
from iseq.typing import MutableStep

TAlphabet = TypeVar("TAlphabet", bound=Alphabet)

HMMER3AltModel = AltModel[NormalState]
HMMER3Fragment = Fragment[TAlphabet, NormalState]
HMMER3Node = Node[NormalState]
HMMER3NullModel = NullModel[NormalState]
HMMER3SearchResults = SearchResults[TAlphabet, NormalState]
HMMER3SpecialNode = SpecialNode[NormalState]
HMMER3Step = Step[MutableStep[NormalState]]

__all__ = [
    "HMMER3AltModel",
    "HMMER3Fragment",
    "HMMER3Node",
    "HMMER3NullModel",
    "HMMER3SearchResults",
    "HMMER3SpecialNode",
    "HMMER3Step",
]



from pathlib import Path
from typing import IO, Optional, Union

from iseq.gff import GFFItem, GFFWriter
from iseq.profile import ProfileID

__all__ = ["OutputWriter"]


class OutputWriter:
    def __init__(self, file: Union[str, Path, IO[str]], item_prefix="item"):
        self._gff = GFFWriter(file)
        self._item_idx = 1
        self._item_prefix = item_prefix

    def write_item(
        self,
        seqid: str,
        seq_alphabet: str,
        profid: ProfileID,
        prof_alphabet: str,
        start: int,
        end: int,
        window_length: int,
        att: Optional[dict] = None,
    ):
        if att is None:
            att = dict()

        item_id = f"{self._item_prefix}{self._item_idx}"
        atts = f"ID={item_id}"
        atts += f";Target_alph={seq_alphabet}"
        atts += f";Profile_name={profid.name}"
        atts += f";Profile_alph={prof_alphabet}"
        atts += f";Profile_acc={profid.acc}"
        atts += f";Window={window_length}"
        for k in sorted(att.keys()):
            atts += f";{k}={att[k]}"

        item = GFFItem(seqid, "iseq", ".", start + 1, end, "0.0", "+", ".", atts)
        self._gff.write_item(item)
        self._item_idx += 1
        return item_id

    def close(self):
        """
        Close the associated stream.
        """
        self._gff.close()

    def __exit__(self, exception_type, exception_value, traceback):
        del exception_type
        del exception_value
        del traceback
        self.close()



import os
import sys
import time
from math import exp

import click
import humanfriendly
import ray
from fasta_reader import FASTAWriter, open_fasta
from hmmer import HMMER
from nmm import DNAAlphabet, Input, IUPACAminoAlphabet, RNAAlphabet
from tqdm import tqdm

from iseq.alphabet import alphabet_name
from iseq.codon_table import CodonTable
from iseq.profile import ProfileID
from iseq.protein import ProteinProfile

from .debug_writer import DebugWriter
from .output_writer import OutputWriter
from .pscan import infer_target_alphabet, update_gff_file


@ray.remote
class Counter:
    def __init__(self):
        self._count = 0

    def increment(self):
        self._count += 1

    def get_and_reset(self):
        count = self._count
        self._count = 0
        return count


@ray.remote
class Worker:
    def __init__(
        self,
        counter: Counter,
        alt_filepath,
        null_filepath,
        target_abc_name: str,
        debug: bool,
    ):
        self._counter = counter
        self._afile = Input.create(alt_filepath)
        self._nfile = Input.create(null_filepath)
        if target_abc_name == "dna":
            target_abc = DNAAlphabet()
        elif target_abc_name == "rna":
            target_abc = RNAAlphabet()
        else:
            raise RuntimeError()
        self._gcode = CodonTable(target_abc, IUPACAminoAlphabet())
        self._output_items = []
        self._codon_seqs = []
        self._amino_seqs = []
        self._debug = debug
        self._debug_table = []
        self._seqids = []

    def set_offset(self, alt_offset, null_offset):
        self._afile.fseek(alt_offset)
        self._nfile.fseek(null_offset)

    def search(self, profids, targets, window):
        # for profid in tqdm(
        #     [ProfileID(*i.split("\t")) for i in profids],
        #     mininterval=10,
        #     desc=self._name,
        # ):
        for profid in iter([ProfileID(*i.split("\t")) for i in profids]):
            alt = self._afile.read()
            null = self._nfile.read()
            prof = ProteinProfile.create_from_binary(profid, null, alt)
            epsilon = prof.epsilon
            prof.window_length = window

            for tgt in targets:
                seq = prof.create_sequence(tgt.sequence.encode())
                search_results = prof.search(seq)
                ifragments = search_results.ifragments()
                seqid = f"{tgt.id}"

                for ifrag in ifragments:
                    start = ifrag.interval.start
                    stop = ifrag.interval.stop
                    self._output_items.append(
                        (
                            seqid,
                            alphabet_name(self._gcode.base_alphabet),
                            prof.profid,
                            alphabet_name(prof.alphabet),
                            start,
                            stop,
                            prof.window_length,
                            {"Epsilon": epsilon},
                        )
                    )
                    codon_frag = ifrag.fragment.decode()
                    self._codon_seqs.append(str(codon_frag.sequence))
                    amino_frag = codon_frag.decode(self._gcode)
                    self._amino_seqs.append(str(amino_frag.sequence))

                if self._debug:
                    self._debug_table += search_results.debug_table()
                self._seqids.append(seqid)

            self._counter.increment.remote()

        self._afile.close()
        self._nfile.close()

    def output_items(self):
        return self._output_items

    def codon_seqs(self):
        return self._codon_seqs

    def amino_seqs(self):
        return self._amino_seqs

    def debug_table(self):
        return self._debug_table

    def seqids(self):
        return self._seqids


def validate_memory(ctx, param, value):
    del ctx
    del param
    if value == "auto":
        return "auto"
    try:
        return str(humanfriendly.parse_size(value))
    except humanfriendly.InvalidSize as e:
        raise click.BadParameter(e.args[0])


@click.command()
@click.argument(
    "profile",
    type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=True),
)
@click.argument("target", type=click.File("r"))
@click.option(
    "--output",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save results to OUTPUT (GFF format).",
    default="output.gff",
)
@click.option(
    "--ocodon",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save codon sequences to OCODON (FASTA format).",
    default="ocodon.fasta",
)
@click.option(
    "--oamino",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save amino acid sequences to OAMINO (FASTA format).",
    default="oamino.fasta",
)
@click.option(
    "--quiet/--no-quiet", "-q/-nq", help="Disable standard output.", default=False,
)
@click.option(
    "--window",
    type=int,
    help="Window length. Defaults to zero, which means no window.",
    default=0,
)
@click.option(
    "--odebug",
    type=click.File("w"),
    help="Save debug info into a tab-separated values file.",
    default=os.devnull,
)
@click.option(
    "--e-value/--no-e-value",
    help="Enable E-value computation. Enabled by default.",
    default=True,
)
@click.option(
    "--ncpus",
    help="Number of CPUs the user wishes to assign to each worker. Defaults to `auto`.",
    default="auto",
    type=str,
)
@click.option(
    "--memory",
    help="The amount of memory that is available for use by workers. You can use units of memory (e.g., 4GB or 100M). Defaults to `auto`.",
    default="auto",
    callback=validate_memory,
    type=str,
)
@click.option(
    "--hit-prefix", help="Hit prefix. Defaults to `item`.", default="item", type=str,
)
def bscan(
    profile,
    target,
    output,
    ocodon,
    oamino,
    quiet,
    window: int,
    odebug,
    e_value: bool,
    ncpus: str,
    memory: str,
    hit_prefix: str,
):
    """
    Binary scan.
    """

    num_cpus: int = 0
    if ncpus == "auto":
        num_cpus = os.cpu_count()
    else:
        num_cpus = int(ncpus)

    alt_offsets = [int(line.strip()) for line in open(profile + ".alt.idx", "r")]
    null_offsets = [int(line.strip()) for line in open(profile + ".null.idx", "r")]

    num_cpus = min(num_cpus, len(alt_offsets))

    kwargs = {}
    if memory != "auto":
        kwargs["memory"] = int(memory)
        kwargs["object_store_memory"] = kwargs["memory"] // 2

    ray.init(
        include_webui=False,
        include_java=False,
        ignore_reinit_error=True,
        num_cpus=num_cpus,
        **kwargs,
    )

    owriter = OutputWriter(output, item_prefix=hit_prefix)
    cwriter = FASTAWriter(ocodon, sys.maxsize)
    awriter = FASTAWriter(oamino, sys.maxsize)
    dwriter = DebugWriter(odebug)

    alt_filepath = (profile + ".alt").encode()
    null_filepath = (profile + ".null").encode()
    meta_filepath = (profile + ".meta").encode()

    profids_list = list(
        split([line.strip() for line in open(meta_filepath, "r")], num_cpus)
    )
    alt_offsets = [i[0] for i in split(alt_offsets, num_cpus)]
    null_offsets = [i[0] for i in split(null_offsets, num_cpus)]

    target_abc = infer_target_alphabet(target)
    if isinstance(target_abc, DNAAlphabet):
        tgt_abc_id = "dna"
    elif isinstance(target_abc, RNAAlphabet):
        tgt_abc_id = "rna"
    else:
        raise RuntimeError()

    with open_fasta(target) as fasta:
        targets = list(fasta)

    workers = []
    i = 0
    counter = Counter.remote()
    for aoffset, noffset, profids in zip(alt_offsets, null_offsets, profids_list):
        debug = odebug is not os.devnull
        w = Worker.remote(counter, alt_filepath, null_filepath, tgt_abc_id, debug)
        w.set_offset.remote(aoffset, noffset)
        w.search.remote(profids, targets, window)
        workers.append(w)
        i += 1

    total = sum(len(profids) for profids in profids_list)
    sleep_time = round(
        max(0.3, 10 - 10 * exp(1 / 10.0) / exp(min(num_cpus / 10.0, 100)))
    )
    with tqdm(total=total, desc="Scan") as pbar:
        while total > 0:
            time.sleep(sleep_time)
            v = ray.get(counter.get_and_reset.remote())
            total -= v
            pbar.update(v)

    for w in workers:
        output_items = ray.get(w.output_items.remote())
        codon_seqs = ray.get(w.codon_seqs.remote())
        amino_seqs = ray.get(w.amino_seqs.remote())
        for item, cseq, aseq in zip(output_items, codon_seqs, amino_seqs):
            item_id = owriter.write_item(*item)
            cwriter.write_item(item_id, cseq)
            awriter.write_item(item_id, aseq)

    if odebug is not os.devnull:
        for w in workers:
            seqids = ray.get(w.seqids.remote())
            debug_table = ray.get(w.debug_table.remote())
            for seqid, debug_row in zip(seqids, debug_table):
                dwriter.write_row(seqid, debug_row)

    owriter.close()
    cwriter.close()
    awriter.close()
    odebug.close_intelligently()

    if e_value:
        hmmer = HMMER(profile)
        result = hmmer.search(oamino, "/dev/null", tblout=True)
        update_gff_file(output, result.tbl)

    ray.shutdown()


def split(a, n):
    k, m = divmod(len(a), n)
    return (a[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n))



import os

import click
from fasta_reader import open_fasta
from hmmer_reader import open_hmmer

from iseq.alphabet import alphabet_name
from iseq.hmmer3 import create_profile
from iseq.hmmer_model import HMMERModel
from iseq.model import EntryDistr

from .debug_writer import DebugWriter
from .output_writer import OutputWriter


@click.command()
@click.argument("profile", type=click.File("r"))
@click.argument("target", type=click.File("r"))
@click.option(
    "--output",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save results to OUTPUT (GFF format).",
    default="output.gff",
)
@click.option(
    "--quiet/--no-quiet", "-q/-nq", help="Disable standard output.", default=False,
)
@click.option(
    "--window",
    type=int,
    help="Window length. Defaults to zero, which means no window.",
    default=0,
)
@click.option(
    "--hmmer3-compat/--no-hmmer3-compat",
    help="Enable full HMMER3 compatibility. Defaults to False.",
    default=False,
)
@click.option(
    "--entry-distr",
    type=click.Choice(["uniform", "occupancy"], case_sensitive=False),
    help="Set the entry distribution. Defaults to occupancy.",
    default="occupancy",
)
@click.option(
    "--odebug",
    type=click.File("w"),
    help="Save debug info into a tab-separated values file.",
    default=os.devnull,
)
def hscan(
    profile,
    target,
    output,
    quiet,
    window: int,
    hmmer3_compat: bool,
    entry_distr: str,
    odebug,
):
    """
    Search nucleotide sequence(s) against a profiles database.

    An OUTPUT line determines an association between a TARGET subsequence and
    a PROFILE protein profile. An association maps a target subsequence to a
    profile and represents a potential homology. Expect many false positive
    associations as we are not filtering out by statistical significance.
    """
    owriter = OutputWriter(output)
    dwriter = DebugWriter(odebug)

    if entry_distr == "occupancy":
        edistr = EntryDistr.OCCUPANCY
    else:
        edistr = EntryDistr.UNIFORM

    if quiet:
        click.open_file(os.devnull, "a")
    else:
        click.get_text_stream("stdout")

    with open_fasta(target) as fasta:
        targets = list(fasta)

    for plain_model in open_hmmer(profile):
        model = HMMERModel(plain_model)
        prof = create_profile(model, hmmer3_compat, edistr, window)
        for tgt in targets:
            seq = prof.create_sequence(tgt.sequence.encode())
            search_results = prof.search(seq)
            ifragments = search_results.ifragments()
            seqid = f"{tgt.id}"
            for interval in [i.interval for i in ifragments]:
                start = interval.start
                stop = interval.stop
                owriter.write_item(
                    seqid,
                    alphabet_name(seq.alphabet),
                    prof.profid,
                    alphabet_name(prof.alphabet),
                    start,
                    stop,
                    prof.window_length,
                )

            if odebug is not os.devnull:
                for i in search_results.debug_table():
                    dwriter.write_row(seqid, i)

    owriter.close()
    odebug.close_intelligently()



import os
from pathlib import Path

from appdirs import user_cache_dir

from .file import make_sure_dir_exist

__all__ = ["ISEQ_CACHE_HOME"]

ISEQ_CACHE_HOME = Path(
    os.environ.get(
        "ISEQ_CACHE_HOME", default=Path(user_cache_dir("iseq", "EBI-Metagenomics")),
    )
)

make_sure_dir_exist(ISEQ_CACHE_HOME)
make_sure_dir_exist(ISEQ_CACHE_HOME / "test_data")
make_sure_dir_exist(ISEQ_CACHE_HOME / "bin")



from nmm import DNAAlphabet, IUPACAminoAlphabet, RNAAlphabet

from iseq.codon_table import CodonTable
from iseq.codon_usage import CodonUsage
from iseq.gencode import GeneticCode
from iseq.random import RandomState


def test_random_codon_dna():

    base_abc = DNAAlphabet()
    amino_abc = IUPACAminoAlphabet()

    table = CodonTable(base_abc, amino_abc, GeneticCode(id=1))
    usage = CodonUsage(table, "9606")

    random = RandomState(0)

    codons = random.codon(usage)
    assert len(codons) == 1
    assert codons[0].symbols == b"GAC"

    codons = random.codon(usage, 3)
    assert len(codons) == 3
    assert codons[0].symbols == b"GGC"
    assert codons[1].symbols == b"GAG"
    assert codons[2].symbols == b"GAC"

    codons = random.codon(usage, 3)
    assert len(codons) == 3
    assert codons[0].symbols == b"CGG"
    assert codons[1].symbols == b"GCC"
    assert codons[2].symbols == b"CTA"

    random = RandomState(0)

    amino_acids = random.amino_acid(usage)
    assert len(amino_acids) == 1
    assert amino_acids[0] == b"D"

    amino_acids = random.amino_acid(usage, 3)
    assert len(amino_acids) == 3
    assert amino_acids[0] == b"G"
    assert amino_acids[1] == b"E"
    assert amino_acids[2] == b"D"

    amino_acids = random.amino_acid(usage, 3)
    assert len(amino_acids) == 3
    assert amino_acids[0] == b"R"
    assert amino_acids[1] == b"A"
    assert amino_acids[2] == b"L"

    amino_acids = random.amino_acid(usage, 500)
    assert amino_acids[442] == b"*"

    amino_acids = random.amino_acid(usage, 1000, include_stop=False)
    assert b"*" not in amino_acids
    assert len(amino_acids) == 1000


def test_random_codon_rna():

    base_abc = RNAAlphabet()
    amino_abc = IUPACAminoAlphabet()

    table = CodonTable(base_abc, amino_abc, GeneticCode(id=1))
    usage = CodonUsage(table, "9606")

    random = RandomState(0)

    codons = random.codon(usage)
    assert len(codons) == 1
    assert codons[0].symbols == b"GAC"

    codons = random.codon(usage, 3)
    assert len(codons) == 3
    assert codons[0].symbols == b"GGC"
    assert codons[1].symbols == b"GAG"
    assert codons[2].symbols == b"GAC"

    codons = random.codon(usage, 3)
    assert len(codons) == 3
    assert codons[0].symbols == b"CGG"
    assert codons[1].symbols == b"GCC"
    assert codons[2].symbols == b"CUA"

    random = RandomState(0)

    amino_acids = random.amino_acid(usage)
    assert len(amino_acids) == 1
    assert amino_acids[0] == b"D"

    amino_acids = random.amino_acid(usage, 3)
    assert len(amino_acids) == 3
    assert amino_acids[0] == b"G"
    assert amino_acids[1] == b"E"
    assert amino_acids[2] == b"D"

    amino_acids = random.amino_acid(usage, 3)
    assert len(amino_acids) == 3
    assert amino_acids[0] == b"R"
    assert amino_acids[1] == b"A"
    assert amino_acids[2] == b"L"

    amino_acids = random.amino_acid(usage, 500)
    assert amino_acids[442] == b"*"

    amino_acids = random.amino_acid(usage, 1000, include_stop=False)
    assert b"*" not in amino_acids
    assert len(amino_acids) == 1000


def test_random_codon_dna_id():

    base_abc = DNAAlphabet()
    amino_abc = IUPACAminoAlphabet()

    table = CodonTable(base_abc, amino_abc, GeneticCode(id=3))
    usage = CodonUsage(table, "8084")

    random = RandomState(0)

    codons = random.codon(usage)
    assert len(codons) == 1
    assert codons[0].symbols == b"GAC"

    codons = random.codon(usage, 3)
    assert len(codons) == 3
    assert codons[0].symbols == b"GGC"
    assert codons[1].symbols == b"GAT"
    assert codons[2].symbols == b"GAC"

    codons = random.codon(usage, 3)
    assert len(codons) == 3
    assert codons[0].symbols == b"CGG"
    assert codons[1].symbols == b"GCC"
    assert codons[2].symbols == b"CTA"

    random = RandomState(0)

    amino_acids = random.amino_acid(usage)
    assert len(amino_acids) == 1
    assert amino_acids[0] == b"D"

    amino_acids = random.amino_acid(usage, 3)
    assert len(amino_acids) == 3
    assert amino_acids[0] == b"G"
    assert amino_acids[1] == b"D"
    assert amino_acids[2] == b"D"

    amino_acids = random.amino_acid(usage, 3)
    assert len(amino_acids) == 3
    assert amino_acids[0] == b"R"
    assert amino_acids[1] == b"A"
    assert amino_acids[2] == b"T"

    amino_acids = random.amino_acid(usage, 5000)
    assert b"*" in amino_acids

    amino_acids = random.amino_acid(usage, 5000, include_stop=False)
    assert b"*" not in amino_acids



import os

from click.testing import CliRunner

from iseq import cli
from iseq.example import example_filepath
from iseq.testing import assert_same_file

_desired_output = """##gff-version 3
Homoserine_dh-consensus	iseq	.	1	519	0.0	+	.	ID=item1;Target_alph=dna;Profile_name=Homoserine_dh;Profile_alph=dna;Profile_acc=PF00742.20;Window=0;Bias=0.2;E-value=2.3e-86;Epsilon=0.01;Score=274.3
AA_kinase-consensus	iseq	.	1	723	0.0	+	.	ID=item2;Target_alph=dna;Profile_name=AA_kinase;Profile_alph=dna;Profile_acc=PF00696.29;Window=0;Bias=4.3;E-value=6.3e-107;Epsilon=0.01;Score=342.7
23ISL-consensus	iseq	.	1	486	0.0	+	.	ID=item3;Target_alph=dna;Profile_name=23ISL;Profile_alph=dna;Profile_acc=PF16620.6;Window=0;Bias=12.1;E-value=4.5e-92;Epsilon=0.01;Score=292.7
"""

_desired_ocodon = """>item1
CCTATCATTTCGACGCTCAAGGAGTCGCTGACAGGTGACCGTATTACTCGAATCGAAGGG
ATATTAAACGGCACCCTGAATTACATTCTCACTGAGATGGAGGAAGAGGGGGCTTCATTC
TCTGAGGCGCTGAAGGAGGCACAGGAATTGGGCTACGCGGAAGCGGATCCTACGGACGAT
GTGGAAGGGCTAGATGCTGCTAGAAAGCTGGCAATTCTAGCCAGATTGGCATTTGGGTTA
GAGGTCGAGTTGGAGGACGTAGAGGTGGAAGGAATTGAAAAGCTGACTGCCGAAGATATT
GAAGAAGCGAAGGAAGAGGGTAAAGTTTTAAAACTAGTGGCAAGCGCCGTCGAAGCCAGG
GTCAAGCCTGAGCTGGTACCTAAGTCACATCCATTAGCCTCGGTAAAAGGCTCTGACAAC
GCCGTGGCTGTAGAAACGGAACGGGTAGGCGAACTCGTAGTGCAGGGACCAGGGGCTGGC
GCAGAGCCAACCGCATCCGCTGTACTCGCTGACCTTCTC
>item2
AAACGTGTAGTTGTAAAGCTTGGGGGTAGTTCTCTGACAGATAAGGAAGAGGCATCACTC
AGGCGTTTAGCTGAGCAGATTGCAGCATTAAAAGAGAGTGGCAATAAACTAGTGGTCGTG
CATGGAGGCGGCAGCTTCACTGATGGTCTGCTGGCATTGAAAAGTGGCCTGAGCTCGGGC
GAATTAGCTGCGGGGTTGAGGAGCACGTTAGAAGAGGCCGGAGAAGTAGCGACGAGGGAC
GCCCTAGCTAGCTTAGGGGAACGGCTTGTTGCAGCGCTGCTGGCGGCGGGTCTCCCTGCT
GTAGGACTCAGCGCCGCTGCGTTAGATGCGACGGAGGCGGGCCGGGATGAAGGCAGCGAC
GGGAACGTCGAGTCCGTGGACGCAGAAGCAATTGAGGAGTTGCTTGAGGCCGGGGTGGTC
CCCGTCCTAACAGGATTTATCGGCTTAGACGAAGAAGGGGAACTGGGAAGGGGATCTTCT
GACACCATCGCTGCGTTACTCGCTGAAGCTTTAGGCGCGGACAAACTCATAATACTGACC
GACGTAGACGGCGTTTACGATGCCGACCCTAAAAAGGTCCCAGACGCGAGGCTCTTGCCA
GAGATAAGTGTGGACGAGGCCGAGGAAAGCGCCTCCGAATTAGCGACCGGTGGGATGAAG
GTCAAACATCCAGCGGCTCTTGCTGCAGCTAGACGGGGGGGTATTCCGGTCGTGATAACG
AAT
>item3
CAGGGTCTGGATAACGCTAATCGTTCGCTAGTTCGCGCTACAAAAGCAGAAAGTTCAGAT
ATACGGAAAGAGGTGACTAACGGCATCGCTAAAGGGCTGAAGCTAGACAGTCTGGAAACA
GCTGCAGAGTCGAAGAACTGCTCAAGCGCACAGAAAGGCGGATCGCTAGCTTGGGCAACC
AACTCCCAACCACAGCCTCTCCGTGAAAGTAAGCTTGAGCCATTGGAAGACTCCCCACGT
AAGGCTTTAAAAACACCTGTGTTGCAAAAGACATCCAGTACCATAACTTTACAAGCAGTC
AAGGTTCAACCTGAACCCCGCGCTCCCGTCTCCGGGGCGCTGTCCCCGAGCGGGGAGGAA
CGCAAGCGCCCAGCTGCGTCTGCTCCCGCTACCTTACCGACACGACAGAGTGGTCTAGGT
TCTCAGGAAGTCGTTTCGAAGGTGGCGACTCGCAAAATTCCAATGGAGTCACAACGCGAG
TCGACT
"""
_desired_oamino = """>item1
PIISTLKESLTGDRITRIEGILNGTLNYILTEMEEEGASFSEALKEAQELGYAEADPTDD
VEGLDAARKLAILARLAFGLEVELEDVEVEGIEKLTAEDIEEAKEEGKVLKLVASAVEAR
VKPELVPKSHPLASVKGSDNAVAVETERVGELVVQGPGAGAEPTASAVLADLL
>item2
KRVVVKLGGSSLTDKEEASLRRLAEQIAALKESGNKLVVVHGGGSFTDGLLALKSGLSSG
ELAAGLRSTLEEAGEVATRDALASLGERLVAALLAAGLPAVGLSAAALDATEAGRDEGSD
GNVESVDAEAIEELLEAGVVPVLTGFIGLDEEGELGRGSSDTIAALLAEALGADKLIILT
DVDGVYDADPKKVPDARLLPEISVDEAEESASELATGGMKVKHPAALAAARRGGIPVVIT
N
>item3
QGLDNANRSLVRATKAESSDIRKEVTNGIAKGLKLDSLETAAESKNCSSAQKGGSLAWAT
NSQPQPLRESKLEPLEDSPRKALKTPVLQKTSSTITLQAVKVQPEPRAPVSGALSPSGEE
RKRPAASAPATLPTRQSGLGSQEVVSKVATRKIPMESQREST
"""

_consensus = """>Homoserine_dh-consensus
CCTATCATTTCGACGCTCAAGGAGTCGCTGACAGGTGACCGTATTACTCGAATCGAAGGG
ATATTAAACGGCACCCTGAATTACATTCTCACTGAGATGGAGGAAGAGGGGGCTTCATTC
TCTGAGGCGCTGAAGGAGGCACAGGAATTGGGCTACGCGGAAGCGGATCCTACGGACGAT
GTGGAAGGGCTAGATGCTGCTAGAAAGCTGGCAATTCTAGCCAGATTGGCATTTGGGTTA
GAGGTCGAGTTGGAGGACGTAGAGGTGGAAGGAATTGAAAAGCTGACTGCCGAAGATATT
GAAGAAGCGAAGGAAGAGGGTAAAGTTTTAAAACTAGTGGCAAGCGCCGTCGAAGCCAGG
GTCAAGCCTGAGCTGGTACCTAAGTCACATCCATTAGCCTCGGTAAAAGGCTCTGACAAC
GCCGTGGCTGTAGAAACGGAACGGGTAGGCGAACTCGTAGTGCAGGGACCAGGGGCTGGC
GCAGAGCCAACCGCATCCGCTGTACTCGCTGACCTTCTC
>AA_kinase-consensus
AAACGTGTAGTTGTAAAGCTTGGGGGTAGTTCTCTGACAGATAAGGAAGAGGCATCACTC
AGGCGTTTAGCTGAGCAGATTGCAGCATTAAAAGAGAGTGGCAATAAACTAGTGGTCGTG
CATGGAGGCGGCAGCTTCACTGATGGTCTGCTGGCATTGAAAAGTGGCCTGAGCTCGGGC
GAATTAGCTGCGGGGTTGAGGAGCACGTTAGAAGAGGCCGGAGAAGTAGCGACGAGGGAC
GCCCTAGCTAGCTTAGGGGAACGGCTTGTTGCAGCGCTGCTGGCGGCGGGTCTCCCTGCT
GTAGGACTCAGCGCCGCTGCGTTAGATGCGACGGAGGCGGGCCGGGATGAAGGCAGCGAC
GGGAACGTCGAGTCCGTGGACGCAGAAGCAATTGAGGAGTTGCTTGAGGCCGGGGTGGTC
CCCGTCCTAACAGGATTTATCGGCTTAGACGAAGAAGGGGAACTGGGAAGGGGATCTTCT
GACACCATCGCTGCGTTACTCGCTGAAGCTTTAGGCGCGGACAAACTCATAATACTGACC
GACGTAGACGGCGTTTACGATGCCGACCCTAAAAAGGTCCCAGACGCGAGGCTCTTGCCA
GAGATAAGTGTGGACGAGGCCGAGGAAAGCGCCTCCGAATTAGCGACCGGTGGGATGAAG
GTCAAACATCCAGCGGCTCTTGCTGCAGCTAGACGGGGGGGTATTCCGGTCGTGATAACG
AAT
>23ISL-consensus
CAGGGTCTGGATAACGCTAATCGTTCGCTAGTTCGCGCTACAAAAGCAGAAAGTTCAGAT
ATACGGAAAGAGGTGACTAACGGCATCGCTAAAGGGCTGAAGCTAGACAGTCTGGAAACA
GCTGCAGAGTCGAAGAACTGCTCAAGCGCACAGAAAGGCGGATCGCTAGCTTGGGCAACC
AACTCCCAACCACAGCCTCTCCGTGAAAGTAAGCTTGAGCCATTGGAAGACTCCCCACGT
AAGGCTTTAAAAACACCTGTGTTGCAAAAGACATCCAGTACCATAACTTTACAAGCAGTC
AAGGTTCAACCTGAACCCCGCGCTCCCGTCTCCGGGGCGCTGTCCCCGAGCGGGGAGGAA
CGCAAGCGCCCAGCTGCGTCTGCTCCCGCTACCTTACCGACACGACAGAGTGGTCTAGGT
TCTCAGGAAGTCGTTTCGAAGGTGGCGACTCGCAAAATTCCAATGGAGTCACAACGCGAG
TCGACT
"""


def test_cli_pscan3(tmp_path):
    os.chdir(tmp_path)
    invoke = CliRunner().invoke
    profile = example_filepath("minifam.hmm")
    with open("consensus.fasta", "w") as file:
        file.write(_consensus)

    with open("desired_output.gff", "w") as file:
        file.write(_desired_output)

    with open("desired_ocodon.fasta", "w") as file:
        file.write(_desired_ocodon)

    with open("desired_oamino.fasta", "w") as file:
        file.write(_desired_oamino)

    r = invoke(cli, ["pscan3", str(profile), "consensus.fasta"])
    assert r.exit_code == 0, r.output
    assert_same_file("oamino.fasta", "desired_oamino.fasta")
    assert_same_file("ocodon.fasta", "desired_ocodon.fasta")
    assert_same_file("output.gff", "desired_output.gff")



"""
GFF3 File Format
----------------

The first line of a GFF3 file must be a comment that identifies the version, e.g.

```
##gff-version 3
```

Fields must be tab-separated. Also, all but the final field in each feature line must contain a
value; "empty" columns should be denoted with a '.'.

- seqid: name of the chromosome or scaffold;
- source: name of the program that generated this feature, or the data source (database or project
  name);
- type: type of feature. Must be a term or accession from the SOFA sequence ontology;
- start: Start position of the feature, with sequence numbering starting at 1;
- end: End position of the feature, with sequence numbering starting at 1;
- score: A floating point value;
- strand: defined as + (forward) or - (reverse);
- phase: One of '0', '1' or '2'. '0' indicates that the first base of the feature is the first base
  of a codon, '1' that the second base is the first base of a codon, and so on;
- attributes: A semicolon-separated list of tag-value pairs, providing additional information about
  each feature. Some of these tags are predefined, e.g. ID, Name, Alias, Parent - see the GFF
  documentation for more details;

Example:

```
##gff-version 3
ctg123 . mRNA            1300  9000  .  +  .  ID=mrna0001;Name=sonichedgehog
ctg123 . exon            1300  1500  .  +  .  ID=exon00001;Parent=mrna0001
ctg123 . exon            1050  1500  .  +  .  ID=exon00002;Parent=mrna0001
ctg123 . exon            3000  3902  .  +  .  ID=exon00003;Parent=mrna0001
ctg123 . exon            5000  5500  .  +  .  ID=exon00004;Parent=mrna0001
ctg123 . exon            7000  9000  .  +  .  ID=exon00005;Parent=mrna0001
```
"""
from __future__ import annotations

import dataclasses
import pathlib
from dataclasses import dataclass
from typing import IO, Any, List, Optional, Tuple, Type, Union

from tqdm.auto import tqdm

__all__ = ["read", "GFF", "GFFItem", "GFFWriter"]


@dataclass
class GFFItem:
    seqid: str
    source: str
    type: str
    start: int
    end: int
    score: str
    strand: str
    phase: str
    attributes: str

    def attributes_astuple(self):
        attrs = []
        for item in self.attributes.split(";"):
            name, value = item.split("=")
            attrs.append((name, value))
        return tuple(attrs)

    def get_attribute(self, name: str):
        return dict(self.attributes_astuple())[name]

    def set_attribute(self, name: str, value: str):
        attrs = self.attributes_astuple()
        new_attrs = []
        found = False
        for n, v in attrs:
            if n == name:
                new_attrs.append((n, value))
                found = True
            else:
                new_attrs.append((n, v))
        if not found:
            raise ValueError(f"Attribute {name} not found.")

        self._set_attributes(new_attrs)

    def _set_attributes(self, attrs: List[Tuple[str, str]]):
        self.attributes = ";".join(f"{n}={v}" for n, v in attrs)

    def __iter__(self):
        return iter(dataclasses.astuple(self))

    @classmethod
    def field_names(cls) -> List[str]:
        return [f.name for f in dataclasses.fields(cls)]

    @classmethod
    def field_types(cls) -> List[Type[Any]]:
        return [f.type for f in dataclasses.fields(cls)]

    def copy(self) -> GFFItem:
        from copy import copy

        return copy(self)


def read(file: Union[str, pathlib.Path, IO[str]], verbose=False) -> GFF:
    from pandas import read_csv

    if isinstance(file, IO):
        close_file = False
    else:
        close_file = True

    if isinstance(file, str):
        file = pathlib.Path(file)

    if isinstance(file, pathlib.Path):
        file = open(file, "r")

    start = file.tell()
    header = file.readline().rstrip()

    names = GFFItem.field_names()
    types = GFFItem.field_types()

    df = read_csv(file, sep="\t", names=names, dtype=dict(zip(names, types)))
    gff = GFF(header)
    total = df.shape[0]
    dis = not verbose
    for _, row in tqdm(df.iterrows(), total=total, desc="GFF", disable=dis, leave=None):
        gff.append(GFFItem(*tuple(row.tolist())))

    if close_file:
        file.close()
    else:
        file.seek(start)

    return gff


class GFF:
    def __init__(self, header: str):
        self._header = header
        self._items: List[GFFItem] = []

    def append(self, item: GFFItem):
        self._items.append(item)

    def filter(self, max_e_value: Optional[float] = None):
        gff = GFF(self._header)

        for item in self.items():
            attrs = explode_attributes(item)

            if max_e_value is not None:
                if "E-value" not in attrs:
                    continue
                e_value = float(attrs["E-value"])
                if e_value > max_e_value:
                    continue

            gff.append(item)

        return gff

    def _to_dataframe(self):
        from pandas import DataFrame

        columns = GFFItem.field_names()
        types = GFFItem.field_types()
        df = DataFrame(self._items, columns=columns, dtype=str)
        for col, typ in zip(columns, types):
            df[col] = df[col].astype(typ)
        return df

    def to_dataframe(self):
        df = self._to_dataframe()
        df = _explode_attributes(df)
        if "att_E-value" in df.columns:
            df["att_E-value"] = df["att_E-value"].astype(float)
        return df

    def _from_dataframe(self, df):
        self._items = []

        keys = GFFItem.field_names()
        for _, row in df.iterrows():
            self.append(GFFItem(**{k: row[k] for k in keys}))

    @property
    def header(self) -> str:
        return self._header

    def deduplicate(self):
        from pandas import concat

        df = self._to_dataframe()
        df = _explode_attributes(df)

        df_stack = []
        for _, df0 in df.groupby(["seqid", "att_Profile"]):
            df_stack.append(_dedup_sequences(df0))
        df = concat(df_stack).sort_index()

        df = _implode_attributes(df)
        self._from_dataframe(df)

    def write_file(self, file: Union[str, pathlib.Path, IO[str]]):
        gff_writer = GFFWriter(file, self._header)
        for item in self.items():
            gff_writer.write_item(item)
        gff_writer.close()

    def items(self) -> List[GFFItem]:
        """
        Get the list of all items.
        """
        return self._items

    def __str__(self) -> str:
        return str(self._to_dataframe())


class GFFWriter:
    def __init__(
        self, file: Union[str, pathlib.Path, IO[str]], header: Optional[str] = None
    ):

        if isinstance(file, str):
            file = pathlib.Path(file)

        if isinstance(file, pathlib.Path):
            file = open(file, "w")

        self._file = file
        if header is None:
            self._file.write("##gff-version 3\n")
        else:
            self._file.write(f"{header}\n")

    def write_item(self, item: GFFItem):
        cols = [
            item.seqid,
            item.source,
            item.type,
            str(item.start),
            str(item.end),
            str(item.score),
            item.strand,
            str(item.phase),
            item.attributes,
        ]
        self._file.write("\t".join(cols))
        self._file.write("\n")

    def close(self):
        """
        Close the associated stream.
        """
        self._file.close()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        del exception_type
        del exception_value
        del traceback
        self.close()


def explode_attributes(item: GFFItem):
    attrs = {}
    for v in item.attributes.split(";"):
        name, value = v.split("=")
        attrs[name] = value
    return attrs


def _explode_attributes(df):
    for i, att in enumerate(df["attributes"]):
        for item in att.split(";"):
            name, value = item.split("=")
            df.at[i, "att_" + name] = value
    return df


def _implode_attributes(df):
    for col in df.columns:
        if col.startswith("att_"):
            del df[col]
    return df


def _dedup_sequences(df):
    df = df.sort_values("start")
    iloc_keep = []
    i = 0
    j = 0
    while i < df.shape[0]:
        start = df.iloc[i]["start"]
        max_end = df.iloc[i]["end"]
        max_end_iloc = i
        j = i + 1
        while j < df.shape[0] and df.iloc[j]["start"] == start:
            end = df.iloc[j]["end"]
            if end > max_end:
                max_end = end
                max_end_iloc = j
            j += 1
        iloc_keep.append(max_end_iloc)
        i = j
        while i < df.shape[0] and df.iloc[i]["end"] <= max_end:
            i += 1
    return df.iloc[iloc_keep]



from typing import Dict

import importlib_resources as pkg_resources
from nmm import Codon, RNAAlphabet

from .codon_table import CodonTable

__all__ = ["CodonUsage"]


class CodonUsage:
    def __init__(self, codon_table: CodonTable, taxonomy_id: str):
        from pandas import read_pickle

        import iseq

        buffer = pkg_resources.open_binary(iseq, "codon_usage_20200706.pkl.xz")

        df = read_pickle(buffer, compression="xz")
        df = df.loc[taxonomy_id]

        if len(df) == 0:
            raise ValueError(f"Unkown taxonomy ID {taxonomy_id}.")

        if isinstance(codon_table.base_alphabet, RNAAlphabet):
            df["Codon"] = df["Codon"].str.replace("T", "U")

        df["Prob"] = df["Number"] / df["Number"].sum()

        self._codon_table = codon_table
        self._map: Dict[Codon, float] = {}

        for _, row in df.iterrows():
            codon = Codon.create(row["Codon"].encode(), codon_table.base_alphabet)
            self._map[codon] = row["Prob"]

    def codon_prob(self, codon: Codon) -> float:
        return self._map[codon]

    def amino_acid_prob(self, amino_acid: bytes) -> float:
        return sum(self.codon_prob(c) for c in self._codon_table.codons(amino_acid))

    @property
    def start_prob(self) -> float:
        return sum(self.codon_prob(c) for c in self._codon_table.start_codons)

    @property
    def stop_prob(self) -> float:
        return sum(self.codon_prob(c) for c in self._codon_table.stop_codons)

    @property
    def codon_table(self) -> CodonTable:
        return self._codon_table



import os
from typing import Dict

import pytest
from click.testing import CliRunner

from iseq import cli
from iseq.testing import assert_same_file

_amino = """>Homoserine_dh-consensus
PIISTLKESLTGDRITRIEGILNGTLNYILTEMEEEGASFSEALKEAQELGYAEADPTDD
VEGLDAARKLAILARLAFGLEVELEDVEVEGIEKLTAEDIEEAKEEGKVLKLVASAVEAR
VKPELVPKSHPLASVKGSDNAVAVETERVGELVVQGPGAGAEPTASAVLADLL
>AA_kinase-consensus
KRVVVKLGGSSLTDKEEASLRRLAEQIAALKESGNKLVVVHGGGSFTDGLLALKSGLSSG
ELAAGLRSTLEEAGEVATRDALASLGERLVAALLAAGLPAVGLSAAALDATEAGRDEGSD
GNVESVDAEAIEELLEAGVVPVLTGFIGLDEEGELGRGSSDTIAALLAEALGADKLIILT
DVDGVYDADPKKVPDARLLPEISVDEAEESASELATGGMKVKHPAALAAARRGGIPVVIT
N
>23ISL-consensus
QGLDNANRSLVRATKAESSDIRKEVTNGIAKGLKLDSLETAAESKNCSSAQKGGSLAWAT
NSQPQPLRESKLEPLEDSPRKALKTPVLQKTSSTITLQAVKVQPEPRAPVSGALSPSGEE
RKRPAASAPATLPTRQSGLGSQEVVSKVATRKIPMESQREST
"""

_nucl_transl: Dict[int, str] = {}

_nucl_transl[
    1
] = """>Homoserine_dh-consensus
CCTATCATTTCGACGCTCAAGGAGTCGCTGACAGGTGACCGTATTACTCGAATCGAAGGG
ATATTAAACGGCACCCTGAATTACATTCTCACTGAGATGGAGGAAGAGGGGGCTTCATTC
TCTGAGGCGCTGAAGGAGGCACAGGAATTGGGCTACGCGGAAGCGGATCCTACGGACGAT
GTGGAAGGGCTAGATGCTGCTAGAAAGCTGGCAATTCTAGCCAGATTGGCATTTGGGTTA
GAGGTCGAGTTGGAGGACGTAGAGGTGGAAGGAATTGAAAAGCTGACTGCCGAAGATATT
GAAGAAGCGAAGGAAGAGGGTAAAGTTTTAAAACTAGTGGCAAGCGCCGTCGAAGCCAGG
GTCAAGCCTGAGCTGGTACCTAAGTCACATCCATTAGCCTCGGTAAAAGGCTCTGACAAC
GCCGTGGCTGTAGAAACGGAACGGGTAGGCGAACTCGTAGTGCAGGGACCAGGGGCTGGC
GCAGAGCCAACCGCATCCGCTGTACTCGCTGACCTTCTC
>AA_kinase-consensus
AAACGTGTAGTTGTAAAGCTTGGGGGTAGTTCTCTGACAGATAAGGAAGAGGCATCACTC
AGGCGTTTAGCTGAGCAGATTGCAGCATTAAAAGAGAGTGGCAATAAACTAGTGGTCGTG
CATGGAGGCGGCAGCTTCACTGATGGTCTGCTGGCATTGAAAAGTGGCCTGAGCTCGGGC
GAATTAGCTGCGGGGTTGAGGAGCACGTTAGAAGAGGCCGGAGAAGTAGCGACGAGGGAC
GCCCTAGCTAGCTTAGGGGAACGGCTTGTTGCAGCGCTGCTGGCGGCGGGTCTCCCTGCT
GTAGGACTCAGCGCCGCTGCGTTAGATGCGACGGAGGCGGGCCGGGATGAAGGCAGCGAC
GGGAACGTCGAGTCCGTGGACGCAGAAGCAATTGAGGAGTTGCTTGAGGCCGGGGTGGTC
CCCGTCCTAACAGGATTTATCGGCTTAGACGAAGAAGGGGAACTGGGAAGGGGATCTTCT
GACACCATCGCTGCGTTACTCGCTGAAGCTTTAGGCGCGGACAAACTCATAATACTGACC
GACGTAGACGGCGTTTACGATGCCGACCCTAAAAAGGTCCCAGACGCGAGGCTCTTGCCA
GAGATAAGTGTGGACGAGGCCGAGGAAAGCGCCTCCGAATTAGCGACCGGTGGGATGAAG
GTCAAACATCCAGCGGCTCTTGCTGCAGCTAGACGGGGGGGTATTCCGGTCGTGATAACG
AAT
>23ISL-consensus
CAGGGTCTGGATAACGCTAATCGTTCGCTAGTTCGCGCTACAAAAGCAGAAAGTTCAGAT
ATACGGAAAGAGGTGACTAACGGCATCGCTAAAGGGCTGAAGCTAGACAGTCTGGAAACA
GCTGCAGAGTCGAAGAACTGCTCAAGCGCACAGAAAGGCGGATCGCTAGCTTGGGCAACC
AACTCCCAACCACAGCCTCTCCGTGAAAGTAAGCTTGAGCCATTGGAAGACTCCCCACGT
AAGGCTTTAAAAACACCTGTGTTGCAAAAGACATCCAGTACCATAACTTTACAAGCAGTC
AAGGTTCAACCTGAACCCCGCGCTCCCGTCTCCGGGGCGCTGTCCCCGAGCGGGGAGGAA
CGCAAGCGCCCAGCTGCGTCTGCTCCCGCTACCTTACCGACACGACAGAGTGGTCTAGGT
TCTCAGGAAGTCGTTTCGAAGGTGGCGACTCGCAAAATTCCAATGGAGTCACAACGCGAG
TCGACT
"""

_nucl_transl[
    9
] = """>Homoserine_dh-consensus
CCTATCATTTCGACGCTCAAGGAGTCCCTCACCGGAGATCGGATAACTCGTATTGAAGGC
ATATTAAACGGCACCCTGAATTACATTCTCACTGAGATGGAGGAAGAGGGGGCTTCATTC
TCTGAGGCGCTGAAGGAGGCGCAAGAGTTAGGCTACGCCGAGGCTGACCCAACTGACGAC
GTAGAGGGACTAGATGCTGCTCGTAAGCTGGCCATATTAGCTCGCCTAGCCTTTGGATTA
GAGGTCGAGTTGGAGGACGTAGAGGTGGAAGGAATTGAAAAGCTCACCGCTGAGGATATT
GAGGAAGCAAAGGAAGAGGGGAAGGTTCTCAAGCTAGTTGCTTCTGCAGTGGAAGCGCGA
GTGAAGCCGGAGCTGGTCCCTAAGTCCCACCCCCTCGCTAGGGTCAAGGGATCTGACAAA
GCTGTAGCTGTCGAGACAGAACGCGTTGGGGAGTTGGTGGTTCAAGGACCGGGAGCGGGG
GCGGAACCCACAGCATCGGCAGTGCTCGCAGATCTCCTA
>AA_kinase-consensus
AAGCGCGTAGTGGTAAAGTTGGGAGGCAGTTCACTCACTGACAAGGAAGAGGCTTCTTTA
CGACGGCTTGCGGAACAAATTGCCGCATTAAAGGAGTCAGGGAAAAAGCTTGTGGTCGTT
CATGGTGGGGGGTCTTTTACAGATGGTCTCCTAGCCTTAAAGAGTGGGCTGTCGAGTGGA
GAGTTGGCCGCGGGTTTACGTAGCACCTTGGAAGAAGCCGGAGAGGTCGCGACCCGTGAT
GCTCTCGCGTCCCTGGGGGAGCGGTTAGTTGCGGCCCTTTTAGCAGCGGGGCTGCCCGCC
GTGGGACTATCTGCCGCTGCGCTTGACGCGACAGAAGCAGGGCGAGATGAGGGCAGGGAC
GGTAATGTTGAAAGAGTGGACGCCGAAGCGATTGAAGAGCTCTTGGAGGCCGGGGTAGTT
CCCGTCCTCACGGGGTTCATCGGCCTCGATGAAGAAGGTGAGTTGGGACGGGGCAGGAGG
GACACCATCGCTGCACTTTTAGCCGAGGCTCTAGGTGCGGATAAGCTGATAATCTTAACA
GATGTCGACGGCGTTTACGATGCGGATCCTAAGAAGGTTCCTGACGCGCGGCTACTCCCA
GAAATCTCCGTCGATGAGGCCGAAGAGTCAGCCAGCGAACTAGCCACCGGAGGGATGAAG
GTGAAGCACCCGGCGGCGTTGGCAGCGGCACGTCGGGGGGGCATCCCCGTAGTCATCACC
AAA
>23ISL-consensus
CAAGGGTTGGATAACGCCAAACGTAGATTAGTACGTGCAACTAAGGCTGAGTCGTCGGAT
ATTCGGAAGGAGGTGACAAATGGGATTGCGAAGGGGCTGAAGCTTGACTCTCTAGAGACT
GCGGCTGAATCCAAGAATTGTAGAAGAGCTCAAAAGGGAGGTAGACTCGCATGGGCGACT
AACAGCCAGCCTCAACCGCTGCGGGAAAGGAAGCTAGAGCCACTTGAAGATTCCCCGCGG
AAGGCGTTGAAGACACCCGTACTCCAAAAGACCTCATCGACGATTACTCTTCAGGCGGTC
AAGGTACAACCCGAACCACGGGCTCCAGTTTCGGGAGCCCTTTCCCCTAGCGGCGAAGAA
CGGAAGCGTCCCGCTGCGTCTGCTCCAGCTACGTTGCCTACGCGACAGAGGGGTTTGGGA
AGTCAAGAAGTAGTCAGGAAGGTTGCTACTCGTAAGATCCCGATGGAAAGACAGCGTGAG
AGCACC
"""

_nucl_transl[
    11
] = """>Homoserine_dh-consensus
CCTATCATTTCGACGCTCAAGGAGTCGCTGACAGGTGACCGTATTACTCGAATCGAAGGG
ATATTAAACGGCACCCTGAATTACATTCTCACTGAGATGGAGGAAGAGGGGGCTTCATTC
TCTGAGGCGCTGAAGGAGGCACAGGAATTGGGCTACGCGGAAGCGGATCCTACGGACGAT
GTGGAAGGGCTAGATGCTGCTAGAAAGCTGGCAATTCTAGCCAGATTGGCATTTGGGTTA
GAGGTCGAGTTGGAGGACGTAGAGGTGGAAGGAATTGAAAAGCTGACTGCCGAAGATATT
GAAGAAGCGAAGGAAGAGGGTAAAGTTTTAAAACTAGTGGCAAGCGCCGTCGAAGCCAGG
GTCAAGCCTGAGCTGGTACCTAAGTCACATCCATTAGCCTCGGTAAAAGGCTCTGACAAC
GCCGTGGCTGTAGAAACGGAACGGGTAGGCGAACTCGTAGTGCAGGGACCAGGGGCTGGC
GCAGAGCCAACCGCATCCGCTGTACTCGCTGACCTTCTC
>AA_kinase-consensus
AAACGTGTAGTTGTAAAGCTTGGGGGTAGTTCTCTGACAGATAAGGAAGAGGCATCACTC
AGGCGTTTAGCTGAGCAGATTGCAGCATTAAAAGAGAGTGGCAATAAACTAGTGGTCGTG
CATGGAGGCGGCAGCTTCACTGATGGTCTGCTGGCATTGAAAAGTGGCCTGAGCTCGGGC
GAATTAGCTGCGGGGTTGAGGAGCACGTTAGAAGAGGCCGGAGAAGTAGCGACGAGGGAC
GCCCTAGCTAGCTTAGGGGAACGGCTTGTTGCAGCGCTGCTGGCGGCGGGTCTCCCTGCT
GTAGGACTCAGCGCCGCTGCGTTAGATGCGACGGAGGCGGGCCGGGATGAAGGCAGCGAC
GGGAACGTCGAGTCCGTGGACGCAGAAGCAATTGAGGAGTTGCTTGAGGCCGGGGTGGTC
CCCGTCCTAACAGGATTTATCGGCTTAGACGAAGAAGGGGAACTGGGAAGGGGATCTTCT
GACACCATCGCTGCGTTACTCGCTGAAGCTTTAGGCGCGGACAAACTCATAATACTGACC
GACGTAGACGGCGTTTACGATGCCGACCCTAAAAAGGTCCCAGACGCGAGGCTCTTGCCA
GAGATAAGTGTGGACGAGGCCGAGGAAAGCGCCTCCGAATTAGCGACCGGTGGGATGAAG
GTCAAACATCCAGCGGCTCTTGCTGCAGCTAGACGGGGGGGTATTCCGGTCGTGATAACG
AAT
>23ISL-consensus
CAGGGTCTGGATAACGCTAATCGTTCGCTAGTTCGCGCTACAAAAGCAGAAAGTTCAGAT
ATACGGAAAGAGGTGACTAACGGCATCGCTAAAGGGCTGAAGCTAGACAGTCTGGAAACA
GCTGCAGAGTCGAAGAACTGCTCAAGCGCACAGAAAGGCGGATCGCTAGCTTGGGCAACC
AACTCCCAACCACAGCCTCTCCGTGAAAGTAAGCTTGAGCCATTGGAAGACTCCCCACGT
AAGGCTTTAAAAACACCTGTGTTGCAAAAGACATCCAGTACCATAACTTTACAAGCAGTC
AAGGTTCAACCTGAACCCCGCGCTCCCGTCTCCGGGGCGCTGTCCCCGAGCGGGGAGGAA
CGCAAGCGCCCAGCTGCGTCTGCTCCCGCTACCTTACCGACACGACAGAGTGGTCTAGGT
TCTCAGGAAGTCGTTTCGAAGGTGGCGACTCGCAAAATTCCAATGGAGTCACAACGCGAG
TCGACT
"""


@pytest.mark.parametrize("trans_tbl", [1, 9, 11])
def test_cli_amino_decode_transl9(tmp_path, trans_tbl):
    os.chdir(tmp_path)
    invoke = CliRunner().invoke

    with open("amino.fa", "w") as file:
        file.write(_amino)

    with open("desired.fa", "w") as file:
        file.write(_nucl_transl[trans_tbl])

    r = invoke(
        cli,
        [
            "amino-decode",
            "amino.fa",
            "--output",
            "nucl.fa",
            "--transl-table",
            trans_tbl,
        ],
    )
    assert r.exit_code == 0, r.output
    assert_same_file("nucl.fa", "desired.fa")



from typing import List

from imm import HMM

from .model import Node, SpecialNode

__all__ = ["core_nodes", "special_node"]

# def transitions(
#     hmm: HMM, nodes: List[Node], special_node: SpecialNode
# ) -> List[Transitions]:
#     breakpoint()
#     MM = hmm.get_transition(nodes[0].M, nodes[1].M)
#     MI = hmm.get_transition(nodes[0].M, nodes[0].I)
#     MD = hmm.get_transition(nodes[0].M, nodes[1].D)
#     IM = hmm.get_transition(nodes[0].I, nodes[1].M)
#     II = hmm.get_transition(nodes[0].I, nodes[0].I)
#     DM = -0.0
#     # DD = -inf
#     pass


def core_nodes(hmm: HMM) -> List[Node]:
    Ms = GrowingList()
    Is = GrowingList()
    Ds = GrowingList()
    for state in hmm.states().values():
        if state.name[:1] == b"M":
            idx = int(state.name[1:]) - 1
            Ms[idx] = state
        elif state.name[:1] == b"I":
            idx = int(state.name[1:]) - 1
            Is[idx] = state
        elif state.name[:1] == b"D":
            idx = int(state.name[1:]) - 1
            Ds[idx] = state

    return [Node(i[0], i[1], i[2]) for i in zip(Ms, Is, Ds)]


def special_node(hmm: HMM) -> SpecialNode:
    S = hmm.find_state(b"S")
    N = hmm.find_state(b"N")
    B = hmm.find_state(b"B")
    E = hmm.find_state(b"E")
    J = hmm.find_state(b"J")
    C = hmm.find_state(b"C")
    T = hmm.find_state(b"T")
    return SpecialNode(S, N, B, E, J, C, T)


class GrowingList(list):
    def __setitem__(self, index, value):
        if index >= len(self):
            self.extend([None] * (index + 1 - len(self)))
        list.__setitem__(self, index, value)



import click
from hmmer_reader import num_models, open_hmmer
from nmm import DNAAlphabet, Model, Output

from iseq.hmmer_model import HMMERModel
from iseq.protein import create_profile


@click.command()
@click.argument(
    "profile",
    type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=True),
)
@click.option(
    "--epsilon", type=float, default=1e-2, help="Indel probability. Defaults to 1e-2."
)
@click.option(
    "--quiet/--no-quiet", "-q/-nq", help="Disable standard output.", default=False,
)
def press(
    profile, epsilon: float, quiet,
):
    """
    Press.

    TODO: fix this doc.
    Search nucleotide sequence(s) against a protein profiles database.

    An OUTPUT line determines an association between a TARGET subsequence and
    a PROFILE protein profile. An association maps a target subsequence to a
    profile and represents a potential homology. Expect many false positive
    associations as we are not filtering out by statistical significance.
    """
    from tqdm import tqdm

    alt_filepath = (profile + ".alt").encode()
    null_filepath = (profile + ".null").encode()
    meta_filepath = (profile + ".meta").encode()

    base_abc = DNAAlphabet()

    if quiet:
        total = 0
    else:
        click.echo("Estimating the number of models... ", nl=False)
        total = num_models(profile)
        click.echo("done.")

    with Output.create(alt_filepath) as afile:
        with Output.create(null_filepath) as nfile:
            with open(meta_filepath, "w") as mfile:
                parser = open_hmmer(profile)
                desc = "Pressing"
                for plain_model in tqdm(parser, total=total, desc=desc, disable=quiet):
                    model = HMMERModel(plain_model)
                    prof = create_profile(model, base_abc, 0, epsilon)

                    hmm = prof.alt_model.hmm
                    dp = hmm.create_dp(prof.alt_model.special_node.T)
                    afile.write(Model.create(hmm, dp))

                    hmm = prof.null_model.hmm
                    dp = hmm.create_dp(prof.null_model.state)
                    nfile.write(Model.create(hmm, dp))

                    name = model.model_id.name
                    acc = model.model_id.acc
                    mfile.write(f"{name}\t{acc}\n")



import pytest


@pytest.fixture
def GALNBKIG_cut():
    from iseq.example import example_filepath

    return {
        "fasta": example_filepath("GALNBKIG_cut.fasta"),
        "gff": example_filepath("PF03373_GALNBKIG_cut.gff"),
        "amino.fasta": example_filepath("PF03373_GALNBKIG_cut.amino.fasta"),
        "codon.fasta": example_filepath("PF03373_GALNBKIG_cut.codon.fasta"),
    }


@pytest.fixture
def large_rna():
    from iseq.example import example_filepath

    return {
        "fasta": example_filepath("large_rna_seq.fasta"),
        "amino0": example_filepath("large_rna_seq_amino0.fasta"),
        "codon0": example_filepath("large_rna_seq_codon0.fasta"),
        "output0": example_filepath("large_rna_seq_output0.gff"),
        "amino48": example_filepath("large_rna_seq_amino48.fasta"),
        "codon48": example_filepath("large_rna_seq_codon48.fasta"),
        "output48": example_filepath("large_rna_seq_output48.gff"),
    }



from __future__ import annotations

from dataclasses import astuple, dataclass
from typing import Callable, Generic, Iterable, List, NamedTuple, Tuple, TypeVar

from imm import Alphabet, Interval, Path, SequenceABC, State, Step

from .fragment import Fragment

__all__ = ["SearchResults", "SearchResult", "create_fragment_type"]

A = TypeVar("A", bound=Alphabet)
S = TypeVar("S", bound=State)

create_fragment_type = Callable[[SequenceABC[A], S, bool], Fragment[A, S]]


@dataclass
class IFragment(Generic[A, S]):
    interval: Interval
    fragment: Fragment[A, S]

    def __iter__(self):
        yield from astuple(self)


DebugRow = NamedTuple(
    "DebugRow",
    [
        ("window_number", int),
        ("window", Interval),
        ("alt_viterbi_score", float),
        ("null_viterbi_score", float),
    ],
)


class SearchResults(Generic[A, S]):
    def __init__(
        self, sequence: SequenceABC[A], create_fragment: create_fragment_type,
    ):
        self._sequence = sequence
        self._create_fragment = create_fragment
        self._results: List[SearchResult[A, S]] = []
        self._windows: List[Interval] = []

    def append(
        self,
        loglik: float,
        window: Interval,
        path: S,
        alt_viterbi_score: float,
        null_viterbi_score: float,
    ):
        subseq = self._sequence[window.start : window.stop]
        r = SearchResult[A, S](
            loglik,
            subseq,
            path,
            self._create_fragment,
            alt_viterbi_score,
            null_viterbi_score,
        )
        self._results.append(r)
        self._windows.append(window)

    @property
    def results(self) -> List[SearchResult[A, S]]:
        return self._results

    @property
    def windows(self) -> List[Interval]:
        return self._windows

    @property
    def length(self) -> int:
        return len(self._results)

    def debug_table(self) -> List[DebugRow]:
        windows = self.windows
        results = self.results
        rows: List[DebugRow] = []
        for win_num, (window, result) in enumerate(zip(windows, results)):

            rows.append(
                DebugRow(
                    win_num,
                    window,
                    result.alt_viterbi_score,
                    result.null_viterbi_score,
                )
            )

        return rows

    def ifragments(self) -> List[IFragment[A, S]]:
        waiting: List[IFragment[A, S]] = []

        windows = self.windows
        results = self.results
        ifragments: List[IFragment[A, S]] = []
        for window, result in zip(windows, results):

            candidates: List[IFragment[A, S]] = []

            for i, frag in zip(result.intervals, result.fragments):
                if not frag.homologous:
                    continue

                interval = Interval(window.start + i.start, window.start + i.stop)
                candidates.append(IFragment(interval, frag))

            ready, waiting = intersect_ifragments(waiting, candidates)
            ifragments.extend(ready)

        ifragments.extend(waiting)

        return ifragments

    def __str__(self) -> str:
        return f"{str(self._results)}"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}:{str(self)}>"


class SearchResult(Generic[A, S]):
    def __init__(
        self,
        loglik: float,
        sequence: SequenceABC[A],
        path: Path[S],
        create_fragment: create_fragment_type,
        alt_viterbi_score: float,
        null_viterbi_score: float,
    ):
        self._loglik = loglik
        self._fragments: List[Fragment[A, S]] = []
        self._intervals: List[Interval] = []
        self._alt_viterbi_score = alt_viterbi_score
        self._null_viterbi_score = null_viterbi_score

        steps = list(path)
        for fragi, stepi, homologous in create_fragments(path):
            substeps = steps[stepi.start : stepi.stop]
            new_steps = [Step.create(s.state, s.seq_len) for s in substeps]
            new_path = Path.create(new_steps)
            seq = sequence[fragi]
            frag = create_fragment(seq, new_path, homologous)
            self._fragments.append(frag)
            self._intervals.append(fragi)

    @property
    def alt_viterbi_score(self) -> float:
        return self._alt_viterbi_score

    @property
    def null_viterbi_score(self) -> float:
        return self._null_viterbi_score

    @property
    def fragments(self) -> List[Fragment[A, S]]:
        return self._fragments

    @property
    def intervals(self) -> List[Interval]:
        return self._intervals

    @property
    def loglikelihood(self) -> float:
        return self._loglik

    def __str__(self) -> str:
        return f"{str(self.loglikelihood)},{str(self._fragments)}"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}:{str(self)}>"


def create_fragments(path: Path) -> Iterable[Tuple[Interval, Interval, bool]]:

    frag_start = frag_stop = 0
    step_start = step_stop = 0
    homologous = False

    for step_stop, step in enumerate(path):

        change = not homologous and step.state.name.startswith(b"M")
        change = change or homologous and step.state.name.startswith(b"E")
        change = change or not homologous and step.state.name.startswith(b"T")

        if change:
            if frag_start < frag_stop:
                fragi = Interval(frag_start, frag_stop)
                stepi = Interval(step_start, step_stop)
                yield (fragi, stepi, homologous)

            frag_start = frag_stop
            step_start = step_stop
            homologous = not homologous

        frag_stop += step.seq_len


def intersect_ifragments(
    waiting: List[IFragment[A, S]], candidates: List[IFragment[A, S]]
) -> Tuple[List[IFragment[A, S]], List[IFragment[A, S]]]:

    ready: List[IFragment[A, S]] = []
    new_waiting: List[IFragment[A, S]] = []

    i = 0
    j = 0

    curr_stop = 0
    while i < len(waiting) and j < len(candidates):

        if waiting[i].interval.start < candidates[j].interval.start:
            ready.append(waiting[i])
            curr_stop = waiting[i].interval.stop
            i += 1
        elif waiting[i].interval.start == candidates[j].interval.start:
            if waiting[i].interval.stop >= candidates[j].interval.stop:
                ready.append(waiting[i])
                curr_stop = waiting[i].interval.stop
            else:
                new_waiting.append(candidates[j])
                curr_stop = candidates[j].interval.stop
            i += 1
            j += 1
        else:
            new_waiting.append(candidates[j])
            curr_stop = candidates[j].interval.stop
            j += 1

        while i < len(waiting) and waiting[i].interval.stop <= curr_stop:
            i += 1

        while j < len(candidates) and candidates[j].interval.stop <= curr_stop:
            j += 1

    while i < len(waiting):
        ready.append(waiting[i])
        i += 1

    while j < len(candidates):
        new_waiting.append(candidates[j])
        j += 1

    return ready, new_waiting



from typing import Optional, Union

from fasta_reader import FASTAParser
from hmmer_reader import HMMERParser
from nmm import BaseAlphabet, DNAAlphabet, IUPACAminoAlphabet, RNAAlphabet

Alphabets = Union[DNAAlphabet, RNAAlphabet, IUPACAminoAlphabet]

__all__ = [
    "Alphabets",
    "alphabet_name",
    "infer_alphabet",
    "infer_fasta_alphabet",
    "infer_hmmer_alphabet",
]


def alphabet_name(alphabet: Alphabets) -> str:
    if isinstance(alphabet, IUPACAminoAlphabet):
        return "amino"
    if isinstance(alphabet, DNAAlphabet):
        return "dna"
    if isinstance(alphabet, RNAAlphabet):
        return "rna"

    # TODO: it is temporary
    if isinstance(alphabet, BaseAlphabet):
        if set(alphabet.symbols) == set(b"ACGT"):
            return "dna"
        if set(alphabet.symbols) == set(b"ACGU"):
            return "rna"
    raise ValueError("Unknown alphabet.")


def infer_alphabet(sequence: bytes) -> Optional[Alphabets]:
    """
    Infer alphabet from a sequence of symbols.

    Parameters
    ----------
    sequence
        Sequence of symbols.
    """
    dna = DNAAlphabet()
    rna = RNAAlphabet()
    amino = IUPACAminoAlphabet()

    abc = set(sequence)

    if len(abc - set(dna.symbols)) == 0:
        return dna

    if len(abc - set(rna.symbols)) == 0:
        return rna

    if len(abc - set(amino.symbols)) == 0:
        return amino

    return None


def infer_fasta_alphabet(parser: FASTAParser) -> Optional[Alphabets]:
    """
    Infer alphabet from fasta file.

    Parameters
    ----------
    parser
        FASTA parser.
    """

    for item in parser:
        alphabet = infer_alphabet(item.sequence.encode())
        if alphabet is not None:
            return alphabet

    return None


def infer_hmmer_alphabet(parser: HMMERParser) -> Optional[Alphabets]:

    for prof in parser:
        alph = dict(prof.metadata)["ALPH"].lower()
        if alph == "amino":
            return IUPACAminoAlphabet()
        if alph == "dna":
            return DNAAlphabet()
        if alph == "rna":
            return RNAAlphabet()

    return None



import os

import pytest
from click.testing import CliRunner

from iseq import cli
from iseq.example import example_filepath
from iseq.testing import assert_same_file


@pytest.mark.slow
def test_cli_pscan_GALNBKIG_pfam100(tmp_path):
    os.chdir(tmp_path)
    invoke = CliRunner().invoke
    profile = example_filepath("Pfam-A.33.1_small.hmm")
    fasta = example_filepath("GALNBKIG_00914_ont_01_plus_strand.fasta")
    oamino = example_filepath("pfam_small_GALNBKIG_00914_oamino.fasta")
    ocodon = example_filepath("pfam_small_GALNBKIG_00914_ocodon.fasta")
    output = example_filepath("pfam_small_GALNBKIG_00914_output.gff")
    r = invoke(cli, ["pscan", str(profile), str(fasta), "--quiet"])
    assert r.exit_code == 0, r.output
    assert_same_file("oamino.fasta", oamino)
    assert_same_file("ocodon.fasta", ocodon)
    assert_same_file("output.gff", output)


def test_cli_pscan_GALNBKIG_00914_ont_01_plus_strand(tmp_path):
    # Related to https://iseq-py.s3.eu-west-2.amazonaws.com/Report+on+NMM+vs+tBLASTn.pdf
    os.chdir(tmp_path)
    invoke = CliRunner().invoke
    profile = example_filepath("PF00113.hmm")
    fasta = example_filepath("GALNBKIG_00914_ont_01_plus_strand.fasta")
    oamino = example_filepath("PF00113_GALNBKIG_00914_ont_01_plus_strand_oamino.fasta")
    ocodon = example_filepath("PF00113_GALNBKIG_00914_ont_01_plus_strand_ocodon.fasta")
    output = example_filepath("PF00113_GALNBKIG_00914_ont_01_plus_strand_output.gff")
    r = invoke(cli, ["pscan", str(profile), str(fasta)])
    assert r.exit_code == 0, r.output
    assert_same_file("oamino.fasta", oamino)
    assert_same_file("ocodon.fasta", ocodon)
    assert_same_file("output.gff", output)


def test_cli_pscan_GALNBKIG_00914_ont_01_plus_strand_window(tmp_path):
    # Related to https://iseq-py.s3.eu-west-2.amazonaws.com/Report+on+NMM+vs+tBLASTn.pdf
    os.chdir(tmp_path)
    invoke = CliRunner().invoke
    profile = example_filepath("PF00113.hmm")
    fasta = example_filepath("GALNBKIG_00914_ont_01_plus_strand.fasta")
    oamino = example_filepath(
        "PF00113_GALNBKIG_00914_ont_01_plus_strand_oamino_window.fasta"
    )
    ocodon = example_filepath(
        "PF00113_GALNBKIG_00914_ont_01_plus_strand_ocodon_window.fasta"
    )
    output = example_filepath(
        "PF00113_GALNBKIG_00914_ont_01_plus_strand_output_window.gff"
    )
    r = invoke(cli, ["pscan", str(profile), str(fasta), "--window", "-1"])
    assert r.exit_code == 0, r.output
    assert_same_file("oamino.fasta", oamino)
    assert_same_file("ocodon.fasta", ocodon)
    assert_same_file("output.gff", output)


def test_cli_pscan_nofile_output(tmp_path, GALNBKIG_cut):
    os.chdir(tmp_path)
    invoke = CliRunner().invoke
    fasta = GALNBKIG_cut["fasta"]
    PF03373 = example_filepath("PF03373.hmm")
    r = invoke(cli, ["pscan", str(PF03373), str(fasta)])
    assert r.exit_code == 0, r.output


def test_cli_pscan_gff_output(tmp_path, GALNBKIG_cut):
    os.chdir(tmp_path)
    PF03373 = example_filepath("PF03373.hmm")
    invoke = CliRunner().invoke
    fasta = GALNBKIG_cut["fasta"]
    output = GALNBKIG_cut["gff"]
    codon = GALNBKIG_cut["codon.fasta"]
    amino = GALNBKIG_cut["amino.fasta"]
    r = invoke(
        cli,
        [
            "pscan",
            str(PF03373),
            str(fasta),
            "--output",
            "output.gff",
            "--ocodon",
            "codon.fasta",
            "--oamino",
            "amino.fasta",
        ],
    )
    assert r.exit_code == 0, r.output
    assert_same_file("output.gff", output)
    assert_same_file("codon.fasta", codon)
    assert_same_file("amino.fasta", amino)


def test_cli_pscan_window0(tmp_path, large_rna):
    os.chdir(tmp_path)
    PF03373 = example_filepath("PF03373.hmm")
    invoke = CliRunner().invoke
    fasta = large_rna["fasta"]
    output = large_rna["output0"]
    codon = large_rna["codon0"]
    amino = large_rna["amino0"]
    r = invoke(
        cli,
        [
            "pscan",
            str(PF03373),
            str(fasta),
            "--output",
            "output.gff",
            "--ocodon",
            "codon.fasta",
            "--oamino",
            "amino.fasta",
            "--window",
            "0",
            "--no-e-value",
        ],
    )
    assert r.exit_code == 0, r.output
    assert_same_file("output.gff", output)
    assert_same_file("codon.fasta", codon)
    assert_same_file("amino.fasta", amino)


def test_cli_pscan_window48(tmp_path, large_rna):
    os.chdir(tmp_path)
    PF03373 = example_filepath("PF03373.hmm")
    invoke = CliRunner().invoke
    fasta = large_rna["fasta"]
    output = large_rna["output48"]
    codon = large_rna["codon48"]
    amino = large_rna["amino48"]
    r = invoke(
        cli,
        [
            "pscan",
            str(PF03373),
            str(fasta),
            "--output",
            "output.gff",
            "--ocodon",
            "codon.fasta",
            "--oamino",
            "amino.fasta",
            "--window",
            "48",
            "--no-e-value",
        ],
    )
    assert r.exit_code == 0, r.output
    assert_same_file("output.gff", output)
    assert_same_file("codon.fasta", codon)
    assert_same_file("amino.fasta", amino)


def test_cli_pscan_large_dataset_window(tmp_path):
    os.chdir(tmp_path)

    profile = example_filepath("PF00113.hmm")
    target = example_filepath("A0ALD9_dna_huge.fasta")
    output = example_filepath("PF00113_A0ALD9_dna_huge_output1776.gff")
    codon = example_filepath("PF00113_A0ALD9_dna_huge_codon1776.fasta")
    amino = example_filepath("PF00113_A0ALD9_dna_huge_amino1776.fasta")

    invoke = CliRunner().invoke
    r = invoke(
        cli,
        [
            "pscan",
            str(profile),
            str(target),
            "--output",
            "output.gff",
            "--ocodon",
            "codon.fasta",
            "--oamino",
            "amino.fasta",
            "--window",
            "1776",
        ],
    )
    assert r.exit_code == 0, r.output
    assert_same_file("output.gff", output)
    assert_same_file("codon.fasta", codon)
    assert_same_file("amino.fasta", amino)


def test_cli_pscan_window0_evalue(tmp_path):
    os.chdir(tmp_path)
    database1 = example_filepath("database1.hmm")
    invoke = CliRunner().invoke
    fasta = example_filepath("PF03373_rna_most_likely.fasta")
    output = example_filepath("PF03373_rna_output.gff")
    oamino = example_filepath("PF03373_rna_oamino.fasta")
    r = invoke(
        cli,
        [
            "pscan",
            str(database1),
            str(fasta),
            "--output",
            "output.gff",
            "--oamino",
            "amino.fasta",
            "--window",
            "0",
            "--e-value",
        ],
    )
    assert r.exit_code == 0, r.output
    assert_same_file("output.gff", output)
    assert_same_file("amino.fasta", oamino)



from typing import TypeVar

from imm import Alphabet
from imm import Fragment as FragmentBase
from imm import Path, SequenceABC, State, Step

__all__ = ["Fragment"]


TState = TypeVar("TState", bound=State)
TAlphabet = TypeVar("TAlphabet", bound=Alphabet)


class Fragment(FragmentBase[TAlphabet, TState]):
    """
    Fragment of a sequence.

    Fragment is a sequence path with homology information.

    Parameters
    ----------
    sequence
        Sequence.
    path
        Path.
    homologous
        Fragment homology.
    """

    def __init__(
        self,
        sequence: SequenceABC[TAlphabet],
        path: Path[Step[TState]],
        homologous: bool,
    ):
        super().__init__(sequence, path)
        self._homologous = homologous

    @property
    def homologous(self) -> bool:
        return self._homologous

    @property
    def sequence(self) -> SequenceABC[TAlphabet]:
        return super().sequence

    @property
    def path(self) -> Path[Step[TState]]:
        return super().path

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}:{str(self)}>"



import os

from click.testing import CliRunner

from iseq import cli
from iseq.example import example_filepath
from iseq.testing import assert_same_file


def test_cli_bscan_GALNBKIG_pfam10(tmp_path):
    os.chdir(tmp_path)
    invoke = CliRunner().invoke
    profile = example_filepath("Pfam-A.33.1_10.hmm")
    fasta = example_filepath("GALNBKIG_00914_ont_01_plus_strand.fasta")

    r = invoke(
        cli,
        [
            "pscan",
            str(profile),
            str(fasta),
            "--output",
            "pscan_output.gff",
            "--ocodon",
            "pscan_ocodon.fasta",
            "--oamino",
            "pscan_oamino.fasta",
            "--quiet",
        ],
    )
    assert r.exit_code == 0, r.output

    r = invoke(cli, ["press", str(profile)])
    assert r.exit_code == 0, r.output

    r = invoke(
        cli,
        [
            "bscan",
            str(profile),
            str(fasta),
            "--output",
            "bscan_output.gff",
            "--ocodon",
            "bscan_ocodon.fasta",
            "--oamino",
            "bscan_oamino.fasta",
            "--quiet",
        ],
    )
    assert r.exit_code == 0, r.output

    assert_same_file("bscan_oamino.fasta", "pscan_oamino.fasta")
    assert_same_file("bscan_ocodon.fasta", "pscan_ocodon.fasta")
    assert_same_file("bscan_output.gff", "pscan_output.gff")


def test_cli_bscan_GALNBKIG_pfam10_ncpus2(tmp_path):
    os.chdir(tmp_path)
    invoke = CliRunner().invoke
    profile = example_filepath("Pfam-A.33.1_10.hmm")
    fasta = example_filepath("GALNBKIG_00914_ont_01_plus_strand.fasta")

    r = invoke(
        cli,
        [
            "pscan",
            str(profile),
            str(fasta),
            "--output",
            "pscan_output.gff",
            "--ocodon",
            "pscan_ocodon.fasta",
            "--oamino",
            "pscan_oamino.fasta",
            "--quiet",
        ],
    )
    assert r.exit_code == 0, r.output

    r = invoke(cli, ["press", str(profile)])
    assert r.exit_code == 0, r.output

    r = invoke(
        cli,
        [
            "bscan",
            str(profile),
            str(fasta),
            "--output",
            "bscan_output.gff",
            "--ocodon",
            "bscan_ocodon.fasta",
            "--oamino",
            "bscan_oamino.fasta",
            "--quiet",
            "--ncpus",
            2,
        ],
    )
    assert r.exit_code == 0, r.output

    assert_same_file("bscan_oamino.fasta", "pscan_oamino.fasta")
    assert_same_file("bscan_ocodon.fasta", "pscan_ocodon.fasta")
    assert_same_file("bscan_output.gff", "pscan_output.gff")



from typing import Union

from imm import MuteState, Step
from nmm import BaseAlphabet, FrameState

from iseq.model import AltModel, Node, NullModel, SpecialNode
from iseq.result import SearchResults

ProteinStep = Step[Union[FrameState, MuteState]]
ProteinSearchResults = SearchResults[BaseAlphabet, FrameState]

ProteinNode = Node[FrameState]
ProteinSpecialNode = SpecialNode[FrameState]
ProteinNullModel = NullModel[FrameState]
ProteinAltModel = AltModel[FrameState]

__all__ = [
    "ProteinStep",
    "ProteinSearchResults",
    "ProteinNode",
    "ProteinSpecialNode",
    "ProteinNullModel",
    "ProteinAltModel",
]



from io import StringIO
from pathlib import Path
from typing import List, NamedTuple, TextIO

import click
from fasta_reader import FASTAItem, FASTAWriter, open_fasta
from hmmer import HMMER
from hmmer_reader import num_models, open_hmmer
from imm import Alphabet
from nmm import IUPACAminoAlphabet
from tqdm import tqdm

from iseq.alphabet import alphabet_name, infer_fasta_alphabet
from iseq.codon_table import CodonTable
from iseq.hmmer_model import HMMERModel
from iseq.profile import ProfileID
from iseq.protein import ProteinProfile, create_profile2

from .output_writer import OutputWriter

HMMEROptions = NamedTuple("HMMEROptions", [("heuristic", bool), ("cut_ga", bool)])


@click.command()
@click.argument(
    "profile",
    type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=False),
)
@click.argument("target", type=click.File("r"))
@click.option(
    "--epsilon", type=float, default=1e-2, help="Indel probability. Defaults to 1e-2."
)
@click.option(
    "--output",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save results to OUTPUT (GFF format).",
    default="output.gff",
)
@click.option(
    "--ocodon",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save codon sequences to OCODON (FASTA format).",
    default="ocodon.fasta",
)
@click.option(
    "--oamino",
    type=click.Path(exists=False, dir_okay=False, writable=True, resolve_path=True),
    help="Save amino acid sequences to OAMINO (FASTA format).",
    default="oamino.fasta",
)
@click.option(
    "--quiet/--no-quiet", "-q/-nq", help="Disable standard output.", default=False,
)
@click.option(
    "--window",
    type=int,
    help="Window length. Defaults to zero, which means no window.",
    default=0,
)
@click.option(
    "--hit-prefix", help="Hit prefix. Defaults to `item`.", default="item", type=str,
)
@click.option(
    "--heuristic/--no-heuristic",
    help="Enable HMMER heuristics. Defaults to True.",
    default=True,
)
@click.option(
    "--cut-ga/--no-cut-ga",
    help="Enable use of profile's GA gathering cutoffs to set all thresholding. Defaults to False.",
    default=False,
)
def pscan3(
    profile: str,
    target: TextIO,
    epsilon: float,
    output: str,
    ocodon: str,
    oamino: str,
    quiet: bool,
    window: int,
    hit_prefix: str,
    heuristic: bool,
    cut_ga: bool,
):
    """
    Search nucleotide sequence(s) against a protein profiles database.

    An OUTPUT line determines an association between a TARGET subsequence and
    a PROFILE protein profile. An association maps a target subsequence to a
    profile and represents a potential homology. Expect many false positive
    associations as we are not filtering out by statistical significance.
    """

    owriter = OutputWriter(output, item_prefix=hit_prefix)
    cwriter = FASTAWriter(ocodon)
    awriter = FASTAWriter(oamino)

    target_abc = infer_target_alphabet(target)

    gcode = CodonTable(target_abc, IUPACAminoAlphabet())
    opts = HMMEROptions(heuristic, cut_ga)
    scan = PScan3(Path(profile), owriter, cwriter, awriter, gcode, opts)
    scan.scan(target, window, epsilon, quiet)
    scan.close()


class PScan3:
    def __init__(
        self,
        profile: Path,
        output: OutputWriter,
        ocodon: FASTAWriter,
        oamino: FASTAWriter,
        codon_table: CodonTable,
        hmmer_options: HMMEROptions,
    ):
        self._profile = profile
        self._output = output
        self._ocodon = ocodon
        self._oamino = oamino
        self._codon_table = codon_table

        hmmer = HMMER(profile)
        hmmer.timeout = 60
        if not hmmer.is_indexed:
            hmmer.index()

        self._hmmer = hmmer
        self._hmmer_options = hmmer_options

    @property
    def num_models(self) -> int:
        return num_models(self._profile)

    def scan(self, target: TextIO, window: int, epsilon: float, quiet: bool):
        with open_fasta(target) as fasta:
            targets = list(fasta)

        base_alphabet = self._codon_table.base_alphabet
        with open_hmmer(self._profile) as pfile:
            total = self.num_models
            profiles = tqdm(iter(pfile), desc="Models", total=total, disable=quiet)
            for plain_model in profiles:
                hmodel = HMMERModel(plain_model)
                prof = create_profile2(hmodel, base_alphabet, window, epsilon)
                self._scan_targets(prof, targets, epsilon, quiet)

    def _scan_targets(
        self,
        prof: ProteinProfile,
        targets: List[FASTAItem],
        epsilon: float,
        quiet: bool,
    ):
        wlen = prof.window_length
        for tgt in tqdm(targets, desc="Targets", leave=False, disable=quiet):
            seq = prof.create_sequence(tgt.sequence.encode())
            search_results = prof.search(seq)
            ifragments = search_results.ifragments()

            tgt_abc = seq.alphabet
            prof_abc = prof.alphabet
            for ifrag in ifragments:
                self._process_fragment(
                    ifrag, f"{tgt.id}", tgt_abc, prof.profid, prof_abc, wlen, epsilon
                )

    def _process_fragment(
        self,
        frag,
        target_id: str,
        target_abc: Alphabet,
        prof_id: ProfileID,
        prof_abc: Alphabet,
        window_length: int,
        epsilon: float,
    ):
        start = frag.interval.start
        stop = frag.interval.stop
        codon_frag = frag.fragment.decode()
        amino_frag = codon_frag.decode(self._codon_table)
        e_value, score, bias = self._target_score(prof_id.acc, str(amino_frag.sequence))
        if score == "NAN":
            return
        item_id = self._output.write_item(
            target_id,
            alphabet_name(target_abc),
            prof_id,
            alphabet_name(prof_abc),
            start,
            stop,
            window_length,
            {"Epsilon": epsilon, "E-value": e_value, "Score": score, "Bias": bias},
        )
        self._ocodon.write_item(item_id, str(codon_frag.sequence))
        self._oamino.write_item(item_id, str(amino_frag.sequence))

    def _target_score(self, acc: str, seq: str):
        result = self._hmmer.search(
            StringIO(">Unknown\n" + seq),
            "/dev/null",
            tblout=True,
            heuristic=self._hmmer_options.heuristic,
            cut_ga=self._hmmer_options.cut_ga,
            hmmkey=acc,
        )

        rows = result.tbl
        if len(rows) == 0:
            e_value = "NAN"
            score = "NAN"
            bias = "NAN"
        else:
            assert len(rows) == 1
            e_value = rows[0].full_sequence.e_value
            score = rows[0].full_sequence.score
            bias = rows[0].full_sequence.bias

        return e_value, score, bias

    def close(self):
        self._output.close()
        self._ocodon.close()
        self._oamino.close()


def infer_target_alphabet(target: TextIO):
    fasta = open_fasta(target)
    target_alphabet = infer_fasta_alphabet(fasta)
    target.seek(0)
    if target_alphabet is None:
        raise click.UsageError("Could not infer alphabet from TARGET.")
    return target_alphabet



from . import gff, hmmer3, protein, testing
from ._cli import cli
from ._testit import test
from ._version import __version__

__all__ = [
    "__version__",
    "cli",
    "gff",
    "hmmer3",
    "protein",
    "test",
    "testing",
]



# It might be useful in the future.
# import click
# from fasta_reader import FASTAWriter, open_fasta

# from iseq.gff import read as read_gff


# @click.command()
# @click.argument("gff", type=click.Path(exists=True, dir_okay=False))
# @click.argument("target", type=click.Path(exists=True, dir_okay=False))
# @click.option(
#     "--gff-output",
#     type=click.Path(exists=False, dir_okay=False),
#     help="Save the surviving GFF items to GFF_OUTPUT.",
# )
# @click.option(
#     "--target-output",
#     type=click.Path(exists=False, dir_okay=False),
#     help="Save the surviving sequences to TARGET_OUTPUT.",
# )
# @click.option(
#     "--e-value", type=float, default=1e-60, help="E-value threshold. Defaults to 1e-60."
# )
# def strip(gff, target, gff_output, target_output, e_value: float):
#     """
#     Filter-out sequences having E-value above a given threshold.
#     """

#     if gff_output is None:
#         gff_output = gff + ".strip"

#     if target_output is None:
#         target_output = target + ".strip"

#     ok_lines = set()
#     ok_items = set()
#     for i, gff_item in enumerate(read_gff(gff).items()):
#         fields = create_fields(gff_item.attributes)
#         if "E-value" in fields:
#             if float(fields["E-value"]) <= e_value:
#                 ok_lines.add(i)
#                 ok_items.add(fields["ID"])

#     with open(gff, "r") as r:
#         with open(gff_output, "w") as w:
#             w.write(r.readline())
#             for i, line in enumerate(r):
#                 if i in ok_lines:
#                     w.write(line)

#     with FASTAWriter(target_output) as writer:
#         with open_fasta(target) as fasta:
#             for tgt in fasta:
#                 if tgt.defline in ok_items:
#                     writer.write_item(tgt.defline, tgt.sequence)

#     print(f"Saved GFF to {gff_output}.")
#     print(f"Saved targets to {target_output}.")


# def create_fields(att: str):
#     fields = {}
#     for v in att.split(";"):
#         name, value = v.split("=", 1)
#         fields[name] = value
#     return fields



from nmm import Codon, DNAAlphabet, IUPACAminoAlphabet, RNAAlphabet

from iseq.codon_table import CodonTable
from iseq.codon_usage import CodonUsage
from iseq.gencode import GeneticCode

EPS = 1e-9


def test_codon_usage_dna_100():
    base_abc = DNAAlphabet()
    amino_abc = IUPACAminoAlphabet()

    table = CodonTable(base_abc, amino_abc)
    usage = CodonUsage(table, "100")

    codon = Codon.create(b"CCT", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0018050541516245488) < EPS
    codon = Codon.create(b"CCC", base_abc)
    assert abs(usage.codon_prob(codon) - 0.016245487364620937) < EPS
    codon = Codon.create(b"CCA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0009025270758122744) < EPS
    codon = Codon.create(b"CCG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.03429602888086643) < EPS

    assert abs(usage.amino_acid_prob(b"P") - 0.05324909747292419) < EPS
    assert abs(usage.amino_acid_prob(b"C") - 0.009927797833935019) < EPS
    assert abs(usage.amino_acid_prob(b"M") - 0.017148014440433214) < EPS

    codon = Codon.create(b"TAG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0009025270758122744) < EPS

    codon = Codon.create(b"TAA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0) < EPS

    codon = Codon.create(b"TGA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.002707581227436823) < EPS

    assert abs(usage.stop_prob - 0.0036101083032490976) < EPS
    assert abs(usage.amino_acid_prob(b"W") - 0.009927797833935019) < EPS

    assert abs(usage.start_prob - 0.058664259927797835) < EPS

    p = sum(usage.amino_acid_prob(aa) for aa in table.amino_acids)
    p += usage.stop_prob
    assert abs(1 - p) < EPS


def test_codon_usage_rna_100():
    base_abc = RNAAlphabet()
    amino_abc = IUPACAminoAlphabet()

    table = CodonTable(base_abc, amino_abc)
    usage = CodonUsage(table, "100")

    codon = Codon.create(b"CCU", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0018050541516245488) < EPS
    codon = Codon.create(b"CCC", base_abc)
    assert abs(usage.codon_prob(codon) - 0.016245487364620937) < EPS
    codon = Codon.create(b"CCA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0009025270758122744) < EPS
    codon = Codon.create(b"CCG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.03429602888086643) < EPS

    assert abs(usage.amino_acid_prob(b"P") - 0.05324909747292419) < EPS
    assert abs(usage.amino_acid_prob(b"C") - 0.009927797833935019) < EPS
    assert abs(usage.amino_acid_prob(b"M") - 0.017148014440433214) < EPS

    codon = Codon.create(b"UAG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0009025270758122744) < EPS

    codon = Codon.create(b"UAA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0) < EPS

    codon = Codon.create(b"UGA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.002707581227436823) < EPS

    assert abs(usage.stop_prob - 0.0036101083032490976) < EPS
    assert abs(usage.amino_acid_prob(b"W") - 0.009927797833935019) < EPS

    assert abs(usage.start_prob - 0.058664259927797835) < EPS

    p = sum(usage.amino_acid_prob(aa) for aa in table.amino_acids)
    p += usage.stop_prob
    assert abs(1 - p) < EPS


def test_codon_usage_dna_100_gencode2():
    base_abc = DNAAlphabet()
    amino_abc = IUPACAminoAlphabet()

    table = CodonTable(base_abc, amino_abc, GeneticCode("Vertebrate Mitochondrial"))
    usage = CodonUsage(table, "100")

    codon = Codon.create(b"CCT", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0018050541516245488) < EPS
    codon = Codon.create(b"CCC", base_abc)
    assert abs(usage.codon_prob(codon) - 0.016245487364620937) < EPS
    codon = Codon.create(b"CCA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0009025270758122744) < EPS
    codon = Codon.create(b"CCG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.03429602888086643) < EPS

    assert abs(usage.amino_acid_prob(b"P") - 0.05324909747292419) < EPS
    assert abs(usage.amino_acid_prob(b"C") - 0.009927797833935019) < EPS
    assert abs(usage.amino_acid_prob(b"M") - 0.018050541516245487) < EPS

    codon = Codon.create(b"TAG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0009025270758122744) < EPS

    codon = Codon.create(b"TAA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0) < EPS

    codon = Codon.create(b"TGA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.002707581227436823) < EPS

    codon = Codon.create(b"AGA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0) < EPS

    codon = Codon.create(b"AGG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.002707581227436823) < EPS

    assert abs(usage.stop_prob - 0.0036101083032490976) < EPS
    assert abs(usage.amino_acid_prob(b"W") - 0.012635379061371842) < EPS

    assert abs(usage.start_prob - 0.07671480144404333) < EPS

    p = sum(usage.amino_acid_prob(aa) for aa in table.amino_acids)
    p += usage.stop_prob
    assert abs(1 - p) < EPS


def test_codon_usage_dna_100_gencode16():
    base_abc = DNAAlphabet()
    amino_abc = IUPACAminoAlphabet()

    table = CodonTable(base_abc, amino_abc, GeneticCode(id=16))
    usage = CodonUsage(table, "100")

    codon = Codon.create(b"CCT", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0018050541516245488) < EPS
    codon = Codon.create(b"CCC", base_abc)
    assert abs(usage.codon_prob(codon) - 0.016245487364620937) < EPS
    codon = Codon.create(b"CCA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0009025270758122744) < EPS
    codon = Codon.create(b"CCG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.03429602888086643) < EPS

    assert abs(usage.amino_acid_prob(b"P") - 0.05324909747292419) < EPS
    assert abs(usage.amino_acid_prob(b"C") - 0.009927797833935019) < EPS
    assert abs(usage.amino_acid_prob(b"M") - 0.017148014440433214) < EPS

    codon = Codon.create(b"TAG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0009025270758122744) < EPS

    codon = Codon.create(b"TAA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0) < EPS

    codon = Codon.create(b"TGA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.002707581227436823) < EPS

    codon = Codon.create(b"AGA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0) < EPS

    codon = Codon.create(b"AGG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.002707581227436823) < EPS

    assert abs(usage.stop_prob - 0.002707581227436823) < EPS
    assert abs(usage.amino_acid_prob(b"W") - 0.009927797833935019) < EPS

    assert abs(usage.start_prob - 0.017148014440433214) < EPS

    p = sum(usage.amino_acid_prob(aa) for aa in table.amino_acids)
    p += usage.stop_prob
    assert abs(1 - p) < EPS


def test_codon_usage_dna_homo_sapiens_gencode1():
    base_abc = DNAAlphabet()
    amino_abc = IUPACAminoAlphabet()

    table = CodonTable(base_abc, amino_abc, GeneticCode(id=1))
    usage = CodonUsage(table, "9606")

    codon = Codon.create(b"CCT", base_abc)
    assert abs(usage.codon_prob(codon) - 0.01754027818499081) < EPS
    codon = Codon.create(b"CCC", base_abc)
    assert abs(usage.codon_prob(codon) - 0.019787725235942962) < EPS
    codon = Codon.create(b"CCA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.016920666769267137) < EPS
    codon = Codon.create(b"CCG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.006924547978778131) < EPS

    assert abs(usage.amino_acid_prob(b"P") - 0.06117321816897904) < EPS
    assert abs(usage.amino_acid_prob(b"C") - 0.023199190843316342) < EPS
    assert abs(usage.amino_acid_prob(b"M") - 0.022035123101626947) < EPS
    assert abs(usage.amino_acid_prob(b"Q") - 0.046575596207835496) < EPS

    codon = Codon.create(b"TAG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0007896448877742195) < EPS

    codon = Codon.create(b"TAA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0009907142640376355) < EPS

    codon = Codon.create(b"TGA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0015551644015129192) < EPS

    codon = Codon.create(b"AGA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.012165533413495484) < EPS

    codon = Codon.create(b"AGG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.011963406553966494) < EPS

    assert abs(usage.stop_prob - 0.003335523553324774) < EPS
    assert abs(usage.amino_acid_prob(b"W") - 0.013171691851737305) < EPS

    assert abs(usage.start_prob - 0.07460160793527573) < EPS

    p = sum(usage.amino_acid_prob(aa) for aa in table.amino_acids)
    p += usage.stop_prob
    assert abs(1 - p) < EPS


def test_codon_usage_dna_homo_sapiens_gencode15():
    base_abc = DNAAlphabet()
    amino_abc = IUPACAminoAlphabet()

    table = CodonTable(base_abc, amino_abc, GeneticCode(id=15))
    usage = CodonUsage(table, "9606")

    codon = Codon.create(b"CCT", base_abc)
    assert abs(usage.codon_prob(codon) - 0.01754027818499081) < EPS
    codon = Codon.create(b"CCC", base_abc)
    assert abs(usage.codon_prob(codon) - 0.019787725235942962) < EPS
    codon = Codon.create(b"CCA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.016920666769267137) < EPS
    codon = Codon.create(b"CCG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.006924547978778131) < EPS

    assert abs(usage.amino_acid_prob(b"P") - 0.06117321816897904) < EPS
    assert abs(usage.amino_acid_prob(b"C") - 0.023199190843316342) < EPS
    assert abs(usage.amino_acid_prob(b"M") - 0.022035123101626947) < EPS
    assert abs(usage.amino_acid_prob(b"Q") - 0.04736524109560972) < EPS

    codon = Codon.create(b"TAG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0007896448877742195) < EPS

    codon = Codon.create(b"TAA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0009907142640376355) < EPS

    codon = Codon.create(b"TGA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.0015551644015129192) < EPS

    codon = Codon.create(b"AGA", base_abc)
    assert abs(usage.codon_prob(codon) - 0.012165533413495484) < EPS

    codon = Codon.create(b"AGG", base_abc)
    assert abs(usage.codon_prob(codon) - 0.011963406553966494) < EPS

    assert abs(usage.stop_prob - 0.0025458786655505545) < EPS
    assert abs(usage.amino_acid_prob(b"W") - 0.013171691851737305) < EPS

    assert abs(usage.start_prob - 0.022035123101626947) < EPS

    p = sum(usage.amino_acid_prob(aa) for aa in table.amino_acids)
    p += usage.stop_prob
    assert abs(1 - p) < EPS



import os

from click.testing import CliRunner

from iseq import cli
from iseq.example import example_filepath
from iseq.testing import assert_same_file


def test_cli_pscan2_pfam24(tmp_path):
    os.chdir(tmp_path)
    invoke = CliRunner().invoke
    profile = example_filepath("Pfam-A_24.hmm")
    fasta = example_filepath("AE014075.1_subset_nucl.fasta")
    oamino = example_filepath("AE014075.1_subset_oamino.fasta")
    ocodon = example_filepath("AE014075.1_subset_ocodon.fasta")
    output = example_filepath("AE014075.1_subset_output.gff")
    r = invoke(
        cli, ["pscan2", str(profile), str(fasta), "--max-e-value", "1e-10", "--quiet"]
    )
    assert r.exit_code == 0, r.output
    assert_same_file("oamino.fasta", oamino)
    assert_same_file("ocodon.fasta", ocodon)
    assert_same_file("output.gff", output)



import click

from .._version import __version__
from ._plot import plot
from .amino_decode import amino_decode
from .bscan import bscan
from .gff_filter import gff_filter
from .hscan import hscan
from .press import press
from .pscan import pscan
from .pscan2 import pscan2
from .pscan3 import pscan3


@click.group(name="iseq", context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(__version__)
def cli():
    """
    Find nucleotide sequences against protein profiles.
    """


cli.add_command(amino_decode)
cli.add_command(bscan)
cli.add_command(gff_filter)
cli.add_command(hscan)
cli.add_command(plot)
cli.add_command(press)
cli.add_command(pscan)
cli.add_command(pscan2)
cli.add_command(pscan3)



from abc import ABC, abstractmethod
from math import log
from typing import Generic, NamedTuple, TypeVar

from imm import Alphabet, Sequence, State, lprob_zero

from .model import AltModel, NullModel, SpecialTransitions
from .result import SearchResults

TAlphabet = TypeVar("TAlphabet", bound=Alphabet)
TState = TypeVar("TState", bound=State)

__all__ = ["Profile", "ProfileID"]

ProfileID = NamedTuple("ProfileID", [("name", str), ("acc", str)])


class Profile(Generic[TAlphabet, TState], ABC):
    def __init__(
        self,
        profid: ProfileID,
        alphabet: Alphabet,
        null_model: NullModel,
        alt_model: AltModel,
        hmmer3_compat: bool,
    ):
        self._profid = profid
        self._alphabet = alphabet
        self._null_model = null_model
        self._alt_model = alt_model
        self._multiple_hits: bool = True
        self._special_trans = SpecialTransitions()
        self._hmmer3_compat = hmmer3_compat
        self._set_target_length_model(1)
        self._window_length: int = 0

    @property
    def profid(self) -> ProfileID:
        return self._profid

    @property
    def window_length(self) -> int:
        return self._window_length

    @window_length.setter
    def window_length(self, length: int) -> None:
        self._window_length = length

    @property
    def alphabet(self):
        return self._alphabet

    @abstractmethod
    def create_sequence(self, sequence: bytes) -> Sequence:
        del sequence
        raise NotImplementedError()

    @property
    def null_model(self) -> NullModel:
        raise NotImplementedError()

    @property
    def alt_model(self) -> AltModel:
        raise NotImplementedError()

    @property
    def multiple_hits(self) -> bool:
        return self._multiple_hits

    @multiple_hits.setter
    def multiple_hits(self, multiple_hits: bool):
        self._multiple_hits = multiple_hits

    @abstractmethod
    def search(self, sequence: Sequence) -> SearchResults[TAlphabet, TState]:
        del sequence
        raise NotImplementedError()

    def _set_target_length_model(self, target_length: int):
        t = self._get_target_length_model(target_length)
        self._null_model.set_special_transitions(t)
        self._alt_model.set_special_transitions(t, self._hmmer3_compat)

    def _get_target_length_model(self, target_length: int) -> SpecialTransitions:
        L = target_length
        if L == 0:
            raise ValueError("Target length cannot be zero.")

        if self._multiple_hits:
            q = 0.5
            log_q = log(0.5)
        else:
            q = 0.0
            log_q = lprob_zero()

        lp = log(L) - log(L + 2 + q / (1 - q))
        l1p = log(2 + q / (1 - q)) - log(L + 2 + q / (1 - q))
        lr = log(L) - log(L + 1)

        t = self._special_trans

        t.NN = t.CC = t.JJ = lp
        t.NB = t.CT = t.JB = l1p
        t.RR = lr
        t.EJ = log_q
        t.EC = log(1 - q)

        return t



from typing import List, TypeVar

from imm import Alphabet, Interval, MuteState, NormalState, Path, Sequence, SequenceABC
from nmm import DNAAlphabet, NTTranslator, NullTranslator, RNAAlphabet

from iseq.hmmer_model import HMMERModel
from iseq.model import EntryDistr, Node, Transitions
from iseq.profile import Profile, ProfileID

from .typing import (
    HMMER3AltModel,
    HMMER3Fragment,
    HMMER3Node,
    HMMER3NullModel,
    HMMER3SearchResults,
    HMMER3SpecialNode,
    HMMER3Step,
)

__all__ = [
    "HMMER3Profile",
    "create_profile",
]


TAlphabet = TypeVar("TAlphabet", bound=Alphabet)


class HMMER3Profile(Profile[TAlphabet, NormalState]):
    def __init__(
        self,
        profid: ProfileID,
        alphabet: TAlphabet,
        null_log_odds: List[float],
        core_nodes: List[Node],
        core_trans: List[Transitions],
        entry_distr: EntryDistr,
        hmmer3_compat=False,
    ):
        R = NormalState.create(b"R", alphabet, null_log_odds)
        null_model = HMMER3NullModel.create(R)

        special_node = HMMER3SpecialNode(
            S=MuteState.create(b"S", alphabet),
            N=NormalState.create(b"N", alphabet, null_log_odds),
            B=MuteState.create(b"B", alphabet),
            E=MuteState.create(b"E", alphabet),
            J=NormalState.create(b"J", alphabet, null_log_odds),
            C=NormalState.create(b"C", alphabet, null_log_odds),
            T=MuteState.create(b"T", alphabet),
        )

        alt_model = HMMER3AltModel.create(
            special_node, core_nodes, core_trans, entry_distr
        )
        super().__init__(profid, alphabet, null_model, alt_model, hmmer3_compat)

        if isinstance(alphabet, (DNAAlphabet, RNAAlphabet)):
            self._translator = NTTranslator()
        else:
            self._translator = NullTranslator()

    @property
    def window_length(self) -> int:
        return super().window_length

    @window_length.setter
    def window_length(self, length: int) -> None:
        if length < -1:
            raise ValueError("Length must be greater than or equal to -1.")

        if length == -1:
            length = 2 * self._alt_model.core_length
        self._window_length = length

    def create_sequence(self, sequence: bytes) -> Sequence:
        seq = self._translator.translate(sequence, self.alphabet)
        return Sequence.create(seq, self.alphabet)

    def search(self, sequence: SequenceABC[TAlphabet]) -> HMMER3SearchResults:

        self._set_target_length_model(len(sequence))

        alt_results = self._alt_model.viterbi(sequence, self.window_length)

        def create_fragment(
            seq: SequenceABC[TAlphabet], path: Path[HMMER3Step], homologous: bool
        ):
            return HMMER3Fragment(seq, path, homologous)

        search_results = HMMER3SearchResults(sequence, create_fragment)

        for alt_result in alt_results:
            subseq = alt_result.sequence
            viterbi_score0 = self._null_model.likelihood(subseq)
            viterbi_score1 = alt_result.loglikelihood
            score = viterbi_score1 - viterbi_score0
            window = Interval(subseq.start, subseq.start + len(subseq))
            if self._hmmer3_compat:
                viterbi_score1 -= 3
            search_results.append(
                score, window, alt_result.path, viterbi_score1, viterbi_score0
            )

        return search_results


def create_profile(
    hmm: HMMERModel,
    hmmer3_compat: bool = False,
    entry_distr: EntryDistr = EntryDistr.OCCUPANCY,
    window_length: int = 0,
) -> HMMER3Profile:
    null_lprobs = hmm.null_lprobs
    null_log_odds = [0.0] * len(null_lprobs)

    nodes: List[HMMER3Node] = []
    for m in range(1, hmm.model_length + 1):
        lodds = [v0 - v1 for v0, v1 in zip(hmm.match_lprobs(m), null_lprobs)]
        M = NormalState.create(f"M{m}".encode(), hmm.alphabet, lodds)
        I = NormalState.create(f"I{m}".encode(), hmm.alphabet, null_log_odds)
        D = MuteState.create(f"D{m}".encode(), hmm.alphabet)

        nodes.append(HMMER3Node(M, I, D))

    trans = hmm.transitions

    profid = ProfileID(hmm.model_id.name, hmm.model_id.acc)
    prof = HMMER3Profile(
        profid, hmm.alphabet, null_log_odds, nodes, trans, entry_distr, hmmer3_compat,
    )
    prof.window_length = window_length
    return prof



from typing import List, Optional

from nmm import Codon

from iseq.codon_usage import CodonUsage

__all__ = ["RandomState"]


class RandomState:
    def __init__(self, seed: Optional[int] = None):
        from numpy.random import RandomState as RndState

        self._random = RndState(seed)

    def codon(
        self, codon_usage: CodonUsage, n: int = 1, include_stop=True
    ) -> List[Codon]:
        codons = codon_usage.codon_table.codons()
        if include_stop:
            p = [codon_usage.codon_prob(c) for c in codons]
        else:
            codons = list(set(codons) - set(codon_usage.codon_table.stop_codons))
            norm = 1 - codon_usage.stop_prob
            p = [codon_usage.codon_prob(c) / norm for c in codons]
        return self._random.choice(codons, n, p=p).tolist()

    def amino_acid(
        self, codon_usage: CodonUsage, n: int = 1, include_stop=True
    ) -> List[bytes]:
        codons = self.codon(codon_usage, n, include_stop)
        return [codon_usage.codon_table.decode(c) for c in codons]



# from numpy.testing import assert_allclose, assert_equal

# from hmmer_reader import open_hmmer
# from iseq.codon import create_profile
# from nmm import GeneticCode
# from nmm.alphabet import IUPACAminoAlphabet, RNAAlphabet
# from nmm.sequence import Sequence


# def test_codon_profile1(PF03373):
#     with open_hmmer(PF03373) as reader:
#         hmmer = create_profile(reader.read_profile(), RNAAlphabet())

#     rna_abc = hmmer.alphabet
#     most_likely_rna_seq = b"CCU GGU AAA GAA GAU AAU AAC AAA".replace(b" ", b"")
#     most_likely_seq = Sequence.create(most_likely_rna_seq, rna_abc)
#     r = hmmer.search(most_likely_seq).results[0]

#     assert_allclose(r.loglikelihood, 125.83363182422178)
#     frags = r.fragments
#     assert_equal(len(frags), 1)
#     frag = frags[0]
#     assert_equal(frag.homologous, True)
#     assert_equal(bytes(frag.sequence), bytes(most_likely_seq))
#     desired = "('CCU', '<M1,3>'),('GGU', '<M2,3>'),('AAA', '<M3,3>')"
#     assert_equal(str(frag)[:53], desired)
